# Copyright Red Hat
#
# snapm/manager/plugins/stratis.py - Snapshot Manager Stratis plugin
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: Apache-2.0
"""
Stratis snapshot manager plugin
"""
from os import stat
from os.path import join as path_join
from subprocess import run, CalledProcessError
from stat import S_ISBLK
from time import time
from uuid import UUID

from dbus.exceptions import DBusException

from dbus_python_client_gen import (
    DPClientInvocationError,
    DPClientSetPropertyContext,
)


from snapm import (
    SnapmCalloutError,
    SnapmBusyError,
    SnapmNoSpaceError,
    SnapmNotFoundError,
    SnapmPluginError,
    SizePolicy,
    SnapStatus,
    Snapshot,
)
from snapm.manager.plugins import (
    DEV_PREFIX,
    DMSETUP_CMD,
    DMSETUP_INFO,
    DMSETUP_NO_HEADINGS,
    DMSETUP_COLUMNS,
    DMSETUP_FIELDS_UUID,
    Plugin,
    parse_snapshot_name,
    device_from_mount_point,
    mount_point_space_used,
    format_snapshot_name,
    encode_mount_point,
)

from .stratislib import (
    StratisdErrors,
    StratisCliStratisdVersionError,
    check_stratisd_version,
    get_object,
    TOP_OBJECT,
    Id,
    IdType,
    MOFilesystem,
    MOPool,
    ObjectManager,
    Pool,
    Filesystem,
    filesystems,
    pools,
)

STRATIS_UUID_PREFIX = "stratis-1-"

STRATIS_DECODE_DM_CMD = "stratis-decode-dm"
STRATIS_DECODE_DM_OUTPUT = "--output"
STRATIS_DECODE_DM_SYMLINK = "symlink"

DBUS_CACHE_VALID = 5

DEV_STRATIS_PREFIX = "/dev/stratis/"

# Minimum allowed Stratis snapshot size (512MiB)
MIN_STRATIS_SNAPSHOT_SIZE = 512 * 1024**2


def is_stratis_device(devpath):
    """
    Test whether ``devpath`` is a Stratis device.

    Return ``True`` if the device at ``devpath`` is a Stratis device or
    ``False`` otherwise.
    """
    dmsetup_cmd_args = [
        DMSETUP_CMD,
        DMSETUP_INFO,
        DMSETUP_COLUMNS,
        DMSETUP_NO_HEADINGS,
        DMSETUP_FIELDS_UUID,
        devpath,
    ]
    try:
        dmsetup_cmd = run(dmsetup_cmd_args, capture_output=True, check=True)
    except CalledProcessError as err:  # pragma: no cover
        if "not found" in err.stderr.decode("utf8"):
            return False
        raise SnapmCalloutError(
            f"Error calling {DMSETUP_CMD}: {err.stderr.decode('utf8')}"
        ) from err
    uuid = dmsetup_cmd.stdout.decode("utf8").strip()
    return uuid.startswith(STRATIS_UUID_PREFIX)


def is_stratisd_running():
    """
    Test whether stratisd is running by attempting to get TOP_OBJECT.

    :returns: ``True`` if the stratisd DBus interface is available or
              ``False`` otherwise.
    """
    try:
        get_object(TOP_OBJECT)
    except DBusException:
        return False
    return True


def stratis_devices_present():
    """
    Test whether any stratis managed devices are present on the system.

    :returns: ``True`` if stratis devices exist or ``False`` otherwise.
    """
    dmsetup_cmd_args = [
        DMSETUP_CMD,
        DMSETUP_INFO,
        DMSETUP_COLUMNS,
        DMSETUP_NO_HEADINGS,
        DMSETUP_FIELDS_UUID,
    ]
    try:
        dmsetup_cmd = run(dmsetup_cmd_args, capture_output=True, check=True)
    except CalledProcessError as err:  # pragma: no cover
        raise SnapmCalloutError(f"Error calling {DMSETUP_CMD}") from err
    for uuid in dmsetup_cmd.stdout.decode("utf8").strip().splitlines():
        if uuid.startswith(STRATIS_UUID_PREFIX):
            return True
    return False


def pool_fs_from_device_path(devpath):
    """
    Return a ``(pool_name, fs_name)`` tuple for the Stratis device at
    ``devpath``.
    """
    stratis_decode_dm_cmd_args = [
        STRATIS_DECODE_DM_CMD,
        STRATIS_DECODE_DM_OUTPUT,
        STRATIS_DECODE_DM_SYMLINK,
        devpath,
    ]
    try:
        stratis_decode_dm_cmd = run(
            stratis_decode_dm_cmd_args, capture_output=True, check=True
        )
    except CalledProcessError as err:  # pragma: no cover
        raise SnapmCalloutError(f"Error calling {STRATIS_DECODE_DM_CMD}") from err
    symlink = stratis_decode_dm_cmd.stdout.decode("utf8").strip()
    return symlink.removeprefix(DEV_STRATIS_PREFIX).split("/")


def pool_fs_from_origin(origin):
    """
    Return a ``(pool_name, fs_name)`` tuple for the Stratis device with
    origin path ``origin``.
    """
    return origin.removeprefix(DEV_STRATIS_PREFIX).split("/")


class StratisSnapshot(Snapshot):
    """
    Class for Stratis snapshot objects.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        name,
        snapset_name,
        origin,
        timestamp,
        mount_point,
        provider,
        pool_name,
        fs_name,
    ):
        super().__init__(name, snapset_name, origin, timestamp, mount_point, provider)
        self.pool_name = pool_name
        self.fs_name = fs_name
        self._pool = None
        self._filesystem = None
        self._cache_ts = 0
        self._get_dbus_cache()

    def __str__(self):
        return "".join(
            [
                super().__str__(),
                f"\nPool:           {self.pool_name}",
                f"\nFilesystem:     {self.fs_name}",
            ]
        )

    @property
    def origin(self):
        return path_join(DEV_PREFIX, "stratis", self.pool_name, self._origin)

    @property
    def origin_options(self):
        """
        File system options needed to specify the origin of this snapshot.

        No file system options are requires for Stratis block snapshots: always
        return the empty string.
        """
        return ""

    @property
    def devpath(self):
        return path_join(DEV_PREFIX, "stratis", self.pool_name, self.fs_name)

    @property
    def status(self):
        (_, filesystem) = self._get_dbus_cache()
        if filesystem.MergeScheduled() is True:
            return SnapStatus.REVERTING
        return SnapStatus.ACTIVE

    @property
    def size(self):
        (_, filesystem) = self._get_dbus_cache()
        return int(filesystem.Size())

    @property
    def free(self):
        (pool, _) = self._get_dbus_cache()
        size = int(pool.TotalPhysicalSize())
        used = int(pool.TotalPhysicalUsed()[1]) if pool.TotalPhysicalUsed()[0] else 0
        return size - used

    # Pylint does not understand the decorator notation.
    # pylint: disable=invalid-overridden-method
    @Snapshot.autoactivate.getter
    def autoactivate(self):
        # Stratis filesystems always autoactivate with the pool
        return True

    def invalidate_cache(self):
        self._pool = None
        self._filesystem = None
        self._cache_ts = 0

    def _get_dbus_cache(self):
        now = time()
        if (
            not (self._pool and self._filesystem)
            or (self._cache_ts + DBUS_CACHE_VALID) < now
        ):
            try:
                proxy = get_object(TOP_OBJECT)
                managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})
            except DBusException as err:
                raise SnapmPluginError(
                    f"Failed to communicate with stratisd: {err}"
                ) from err

            (pool, filesystem) = _get_pool_filesystem(
                managed_objects, self.pool_name, self.fs_name
            )
            self._pool = pool
            self._filesystem = filesystem
            self._cache_ts = time()
        return (self._pool, self._filesystem)


def filter_stratis_snapshot(filesystem):
    """
    Filter Stratis snapshots.

    Return ``True`` if the filesystem epresented by ``filesystem`` is a stratis
    snapshot or ``False`` otherwise. The ``filesystem`` argument must be a
    DBus managed object corresponding to a Stratis filesystem.
    """
    return filesystem.Origin()[0]


def _snapshot_min_size(policy_size):
    """
    Return the minimum snapshot size given the space used by the snapshot
    mount point.

    :param policy_size: The size suggested by the in-use size policy.
    :returns: The greater of ``policy_size`` and ``MIN_STRATIS_SNAPSHOT_SIZE``.
    """
    return max(MIN_STRATIS_SNAPSHOT_SIZE, policy_size)


def _get_pool_filesystem(managed_objects, pool_name, fs_name):
    """
    Return pool and filesystem managed objects.

    :param managed_objects: The managed objects to search.
    :param pool_name: The name of the pool to return.
    :param fs_name: The name of the filesystem to return, or ``None``
                    to query only the pool.
    """
    props = {"Name": pool_name}
    pool_object_path = next(
        pools(props=props).require_unique_match(True).search(managed_objects)
    )[0]

    if fs_name is not None:
        fs_id = Id(IdType.NAME, fs_name)
        fs_props = {"Pool": pool_object_path} | fs_id.managed_objects_key()

        filesystem = [
            MOFilesystem(info)
            for objpath, info in filesystems(props=fs_props)
            .require_unique_match(True)
            .search(managed_objects)
        ][0]
    else:
        filesystem = None

    pool = [
        MOPool(info)
        for objpath, info in pools(props=props)
        .require_unique_match(True)
        .search(managed_objects)
    ][0]
    return (pool, filesystem)


def _origin_uuid_to_fs_name(managed_objects, pool_object_path, origin_uuid):
    """
    Return the filesystem corresponding to `origin_uuid`.
    :param managed_objects: DBus managed objects for the service
    :param pool_object_path: The pool DBus object path
    :param origin_uuid: The origin UUID to find.
    """
    fs_id = Id(IdType.UUID, UUID(origin_uuid))
    fs_props = {"Pool": pool_object_path} | fs_id.managed_objects_key()

    filesystem = [
        MOFilesystem(info)
        for objpath, info in filesystems(props=fs_props)
        .require_unique_match(True)
        .search(managed_objects)
    ][0]

    return str(filesystem.Name())


def _fs_name_to_uuid(managed_objects, pool_object_path, fs_name):
    """
    Return the filesystem UUID corresponding to `origin_name`.
    :param managed_objects: DBus managed objects for the service
    :param pool_object_path: The pool DBus object path
    :param origin_name: The origin filesystem name to find
    """
    fs_props = {"Pool": pool_object_path, "Name": fs_name}

    filesystem = [
        MOFilesystem(info)
        for objpath, info in filesystems(props=fs_props)
        .require_unique_match(True)
        .search(managed_objects)
    ][0]

    return str(filesystem.Uuid())


def _find_in_progress_merge(managed_objects, pool_object_path, origin_uuid):
    """
    Return a list containing any in-progres merge for the specified
    `pool_object_path` and `origin_uuid`, or the empty list if no merge is
    in progress.
    :param managed_objects: DBus managed objects for the service
    :param pool_object_path: The pool DBus object path
    :param origin_uuid: The origin filesystem uuid to find
    """
    fs_props = {
        "Pool": pool_object_path,
        "Origin": (True, origin_uuid),
        "MergeScheduled": True,
    }

    return [
        MOFilesystem(info)
        for objpath, info in filesystems(props=fs_props).search(managed_objects)
    ]


def _pool_free_space_bytes(managed_objects, pool_name):
    """
    Return the free space available as bytes for the Stratis pool named
    ``pool_name``.
    """
    (pool, _) = _get_pool_filesystem(managed_objects, pool_name, None)
    size = int(pool.TotalPhysicalSize())
    used = int(pool.TotalPhysicalUsed()[1]) if pool.TotalPhysicalUsed()[0] else 0
    return size - used


def _fs_size_bytes(managed_objects, pool_name, fs_name):
    """
    Return the size of the specified filesystem in bytes.
    """
    (_, filesystem) = _get_pool_filesystem(managed_objects, pool_name, fs_name)
    return int(filesystem.Size())


class Stratis(Plugin):
    """
    Class for Stratis snapshot plugin.
    """

    name = "stratis"
    version = "0.1.0"
    snapshot_class = StratisSnapshot

    def __init__(self, logger):
        """
        Initialise the Stratis plugin.

        :param logger: The logger to pass to the Plugin class.
        """
        super().__init__(logger)
        try:
            check_stratisd_version()
        except DBusException as err:
            if stratis_devices_present():
                self._log_warn(
                    "Stratis devices present but stratisd is not running: %s", err
                )
                self._log_warn(
                    "Run 'systemctl enable --now stratisd' to manage snapshots for Stratis volumes"
                )
            raise SnapmNotFoundError(
                f"Stratisd DBus service not available: {err}"
            ) from err
        except StratisCliStratisdVersionError as err:
            raise SnapmPluginError(f"Stratisd version check failed: {err}") from err

    # pylint: disable=too-many-locals
    def discover_snapshots(self):
        """
        Discover snapshots managed by this plugin class.

        Returns a list of objects that are a subclass of ``Snapshot``.
        """
        snapshots = []

        try:
            proxy = get_object(TOP_OBJECT)
            managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})
        except DBusException as err:
            raise SnapmPluginError(
                f"Failed to communicate with stratisd: {err}"
            ) from err

        path_to_name = dict(
            (path, MOPool(info).Name())
            for path, info in pools().search(managed_objects)
        )

        filesystems_with_props = [
            MOFilesystem(info)
            for objpath, info in filesystems(props=None)
            .require_unique_match(False)
            .search(managed_objects)
        ]

        for filesystem in filesystems_with_props:
            if not filter_stratis_snapshot(filesystem):
                continue

            pool_name = path_to_name[filesystem.Pool()]
            filesystem_name = str(filesystem.Name())

            origin = _origin_uuid_to_fs_name(
                managed_objects, filesystem.Pool(), str(filesystem.Origin()[1])
            )

            try:
                fields = parse_snapshot_name(filesystem_name, origin)
            except ValueError:
                continue
            if fields is not None:
                (snapset, timestamp, mount_point) = fields
                full_name = f"{pool_name}/{filesystem_name}"
                self._log_debug("Found %s snapshot: %s", self.name, full_name)
                snapshots.append(
                    StratisSnapshot(
                        full_name,
                        snapset,
                        origin,
                        timestamp,
                        mount_point,
                        self,
                        pool_name,
                        filesystem_name,
                    )
                )

        return snapshots

    def can_snapshot(self, source):
        """
        Test whether this plugin can snapshot the specified mount point.

        :param source: The mount point path to test.
        :returns: ``True`` if this plugin can snapshot the file system mounted
                  at ``mount_point``, or ``False`` otherwise.
        """
        if S_ISBLK(stat(source).st_mode):
            device = source
        else:
            device = device_from_mount_point(source)

        if not is_stratis_device(device):
            return False
        if not is_stratisd_running():
            self._log_error("Stratis source specified but stratisd is not running")
            return False
        return True

    # pylint: disable=too-many-arguments
    def _check_free_space(self, managed_objects, origin, mount_point, size_policy):
        """
        Check for available space in pool ``pool_name`` for the specified
        mount point.

        :param pool_name: The name of the pool to check.
        :param fs_name: The name of the filesystem to check.
        :param mount_point: The mount point path to check.
        :param size_policy: The size policy to be applied.
        :returns: The minimum size required for the snapshot.
        :raises: ``SnapmNoSpaceError`` if the minimum snapshot size exceeds the
                 available space.
        """
        pool_name, fs_name = pool_fs_from_origin(origin)
        fs_used = mount_point_space_used(mount_point)
        pool_free = _pool_free_space_bytes(managed_objects, pool_name)
        fs_size = _fs_size_bytes(managed_objects, pool_name, fs_name)
        policy = SizePolicy(
            origin, mount_point, pool_free, fs_used, fs_size, size_policy
        )
        snapshot_min_size = _snapshot_min_size(policy.size)
        if pool_free < (sum(self.size_map[pool_name].values()) + snapshot_min_size):
            raise SnapmNoSpaceError(
                f"Stratis pool {pool_name} has insufficient free space to snapshot {mount_point}"
            )
        return snapshot_min_size

    # pylint: disable=too-many-arguments
    def check_create_snapshot(
        self, origin, snapset_name, timestamp, mount_point, size_policy
    ):
        """
        Perform pre-creation checks before creating a snapshot.

        :param origin: The origin volume for the snapshot.
        :param snapset_name: The name of the snapshot set to be created.
        :param timestamp: The snapshot set timestamp.
        :param mount_point: The mount point path for this snapshot.
        :raises: ``SnapmNoSpaceError`` if there is insufficient free space to
                 create the snapshot.
        """
        pool_name, fs_name = pool_fs_from_origin(origin)
        try:
            proxy = get_object(TOP_OBJECT)
            managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})
        except DBusException as err:
            raise SnapmPluginError(
                f"Failed to communicate with stratisd: {err}"
            ) from err

        if pool_name not in self.size_map:
            self.size_map[pool_name] = {}
            self.size_map[pool_name][fs_name] = self._check_free_space(
                managed_objects, origin, mount_point, size_policy
            )

    # pylint: disable=too-many-arguments
    def create_snapshot(
        self, origin, snapset_name, timestamp, mount_point, size_policy
    ):
        """
        Create a snapshot of ``origin`` in the snapset named ``snapset_name``.

        :param origin: The origin volume for the snapshot.
        :param snapset_name: The name of the snapshot set to be created.
        :param timestamp: The snapshot set timestamp.
        :param mount_point: The mount point path for this snapshot.
        :raises: ``SnapmNoSpaceError`` if there is insufficient free space to
                 create the snapshot.
        """
        pool_name, fs_name = pool_fs_from_origin(origin)

        snapshot_name = format_snapshot_name(
            fs_name, snapset_name, timestamp, encode_mount_point(mount_point)
        )

        try:
            proxy = get_object(TOP_OBJECT)
            managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})
        except DBusException as err:
            raise SnapmPluginError(
                f"Failed to communicate with stratisd: {err}"
            ) from err

        self._check_free_space(managed_objects, origin, mount_point, size_policy)

        self._log_debug(
            "Creating Stratis snapshot for %s/%s mounted at %s",
            pool_name,
            fs_name,
            mount_point,
        )
        (pool_object_path, _) = next(
            pools(props={"Name": pool_name})
            .require_unique_match(True)
            .search(managed_objects)
        )
        (origin_fs_object_path, _) = next(
            filesystems(props={"Name": fs_name, "Pool": pool_object_path})
            .require_unique_match(True)
            .search(managed_objects)
        )

        ((changed, _), return_code, message) = Pool.Methods.SnapshotFilesystem(
            get_object(pool_object_path),
            {"origin": origin_fs_object_path, "snapshot_name": snapshot_name},
        )

        if return_code != StratisdErrors.OK:  # pragma: no cover
            raise SnapmPluginError(message)

        if not changed:  # pragma: no cover
            raise SnapmPluginError(
                f"Stratis daemon reported no change creating snapshot {snapshot_name}"
            )

        return StratisSnapshot(
            f"{pool_name}/{snapshot_name}",
            snapset_name,
            fs_name,
            timestamp,
            mount_point,
            self,
            pool_name,
            snapshot_name,
        )

    def origin_from_mount_point(self, mount_point):
        """
        Return a string representing the origin from a given mount point path.
        """
        device = device_from_mount_point(mount_point)
        if not is_stratis_device(device):
            return None
        (pool_name, fs_name) = pool_fs_from_device_path(device)
        return path_join(DEV_STRATIS_PREFIX, pool_name, fs_name)

    def delete_snapshot(self, name):
        """
        Delete the snapshot named ``name``

        :param name: The name of the snapshot to be removed.
        """
        (pool_name, fs_name) = name.split("/")
        fs_name = [fs_name]

        try:
            proxy = get_object(TOP_OBJECT)
            managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})
        except DBusException as err:
            raise SnapmPluginError(
                f"Failed to communicate with stratisd: {err}"
            ) from err

        (pool_object_path, _) = next(
            pools(props={"Name": pool_name})
            .require_unique_match(True)
            .search(managed_objects)
        )

        requested_names = frozenset(fs_name)

        pool_filesystems = {
            MOFilesystem(info).Name(): op
            for (op, info) in filesystems(props={"Pool": pool_object_path}).search(
                managed_objects
            )
        }
        already_removed = requested_names.difference(frozenset(pool_filesystems.keys()))

        if already_removed != frozenset():  # pragma: no cover
            raise SnapmPluginError(
                f"Stratisd destroy reported snapshot already removed: {already_removed}"
            )

        fs_object_paths = [
            op for (name, op) in pool_filesystems.items() if name in requested_names
        ]

        (
            (destroyed, list_destroyed),
            return_code,
            message,
        ) = Pool.Methods.DestroyFilesystems(
            get_object(pool_object_path), {"filesystems": fs_object_paths}
        )

        if return_code != StratisdErrors.OK:  # pragma: no cover
            raise SnapmPluginError(message)

        if not destroyed or len(list_destroyed) < len(
            fs_object_paths
        ):  # pragma: no cover
            raise SnapmPluginError(
                (
                    f"Expected to destroy the specified filesystems in pool "
                    f"{pool_name} but stratisd reports that it did not"
                    f"actually destroy some or all of the filesystems "
                    f"requested"
                )
            )

    # pylint: disable=too-many-arguments
    def rename_snapshot(self, old_name, origin, snapset_name, timestamp, mount_point):
        """
        Rename the snapshot named ``old_name`` according to the provided
        snapshot field values.

        :param old_name: The original name of the snapshot to be renamed.
        :param origin: The origin volume for the snapshot.
        :param snapset_name: The new name of the snapshot set.
        :param timestamp: The snapshot set timestamp.
        :param mount_point: The mount point of the snapshot.
        """
        (pool_name, fs_name) = old_name.split("/")
        _, origin = pool_fs_from_origin(origin)
        new_name = format_snapshot_name(
            origin, snapset_name, timestamp, encode_mount_point(mount_point)
        )

        self._log_debug(
            "Renaming snapshot from %s to %s",
            old_name,
            new_name,
        )

        try:
            proxy = get_object(TOP_OBJECT)
            managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})
        except DBusException as err:
            raise SnapmPluginError(
                f"Failed to communicate with stratisd: {err}"
            ) from err

        (pool_object_path, _) = next(
            pools(props={"Name": pool_name})
            .require_unique_match(True)
            .search(managed_objects)
        )
        (fs_object_path, _) = next(
            filesystems(props={"Name": fs_name, "Pool": pool_object_path})
            .require_unique_match(True)
            .search(managed_objects)
        )
        ((changed, _), return_code, message) = Filesystem.Methods.SetName(
            get_object(fs_object_path), {"name": new_name}
        )

        if return_code != StratisdErrors.OK:  # pragma: no cover
            raise SnapmPluginError(
                f"Rename failed: stratisd returned error {return_code}, {message})"
            )

        if not changed:
            raise SnapmPluginError(f"Rename reported no change: {new_name}")

        return StratisSnapshot(
            f"{pool_name}/{new_name}",
            snapset_name,
            fs_name,
            timestamp,
            mount_point,
            self,
            pool_name,
            new_name,
        )

    def check_resize_snapshot(self, name, origin, mount_point, size_policy):
        """
        Check whether this snapshot can be resized or not. This method returns
        if the current snapshot can be resized and raises an exception if not.

        :returns: None
        :raises: ``SnapmNoSpaceError`` if there is insufficient space to resize
                 the snapshot according to ``size_policy`` or ``SnapmPluginError``
                 if another error occurs.
        """
        pool_name, fs_name = pool_fs_from_origin(origin)
        try:
            proxy = get_object(TOP_OBJECT)
            managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})
        except DBusException as err:
            raise SnapmPluginError(
                f"Failed to communicate with stratisd: {err}"
            ) from err

        if pool_name not in self.size_map:
            self.size_map[pool_name] = {}
            self.size_map[pool_name][fs_name] = self._check_free_space(
                managed_objects, origin, mount_point, size_policy
            )

    def resize_snapshot(self, name, origin, mount_point, size_policy):
        """
        Perform any necessary resize operation on this snapshot. Since Stratis
        snapshots use space allocated dynamically from the thin pool this is a
        no-op for Stratis devices.
        """
        return

    def check_revert_snapshot(self, name, origin):
        """
        Check whether this snapshot can be reverted or not. This method returns
        if the current snapshot can be reverted and raises an exception if not.

        :returns: None
        :raises: ``NotImplementedError`` if this plugin does not support the
        revert operation, ``SnapmBusyError`` if the snapshot is already in the
        process of being reverted to another snapshot state or
        ``SnapmPluginError`` if another reason prevents the snapshot from being
        merged.
        """
        pool_name, origin = pool_fs_from_origin(origin)

        try:
            proxy = get_object(TOP_OBJECT)
            managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})
        except DBusException as err:
            raise SnapmPluginError(
                f"Failed to communicate with stratisd: {err}"
            ) from err

        (pool_object_path, _) = next(
            pools(props={"Name": pool_name})
            .require_unique_match(True)
            .search(managed_objects)
        )

        origin_uuid = _fs_name_to_uuid(managed_objects, pool_object_path, origin)
        in_progress = _find_in_progress_merge(
            managed_objects, pool_object_path, origin_uuid
        )

        if len(in_progress):
            raise SnapmBusyError(
                f"Snapshot revert is in progress for {name} origin volume {pool_name}/{origin}"
            )

    def revert_snapshot(self, name):
        """
        Revert the state of the content of the origin to the content at the
        time the snapshot was taken.

        For Stratis snapshots of in-use filesystems this will take place at
        the next activation (typically a reboot into the revert boot entry
        for the snapshot set).
        """
        (pool_name, fs_name) = name.split("/")

        try:
            proxy = get_object(TOP_OBJECT)
            managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})
        except DBusException as err:
            raise SnapmPluginError(
                f"Failed to communicate with stratisd: {err}"
            ) from err

        (pool_object_path, _) = next(
            pools(props={"Name": pool_name})
            .require_unique_match(True)
            .search(managed_objects)
        )
        (fs_object_path, info) = next(
            filesystems(props={"Name": fs_name, "Pool": pool_object_path})
            .require_unique_match(True)
            .search(managed_objects)
        )
        filesystem = MOFilesystem(info)

        try:
            Filesystem.Properties.MergeScheduled.Set(get_object(fs_object_path), True)
        except DPClientInvocationError as err:
            if isinstance(err.context, DPClientSetPropertyContext):
                origin_uuid = filesystem.Origin()[1]
                if len(
                    _find_in_progress_merge(
                        managed_objects, pool_object_path, origin_uuid
                    )
                ):
                    origin = _origin_uuid_to_fs_name(
                        managed_objects, pool_object_path, origin_uuid
                    )
                    raise SnapmBusyError(
                        f"Snapshot revert is in progress for {fs_name} origin volume "
                        f"{pool_name}/{origin}"
                    ) from err
            raise SnapmPluginError(
                f"Unexpected D-Bus error setting property for {pool_name}/{fs_name}"
            ) from err

    def activate_snapshot(self, name):
        """
        Activate the snapshot named ``name``

        :param name: The name of the snapshot to be activated.
        """
        return

    def deactivate_snapshot(self, name):
        """
        Deactivate the snapshot named ``name``

        :param name: The name of the snapshot to be deactivated.
        """
        return

    def set_autoactivate(self, name, auto=False):
        """
        Set the autoactivation state of the snapshot named ``name``.

        :param name: The name of the snapshot to be modified.
        :param auto: ``True`` to enable autoactivation or ``False`` otherwise.
        """
        return



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Package mediating dbus actions.
"""

from ._stratisd_version import check_stratisd_version
from ._stratisd_constants import StratisdErrors
from ._errors import (
    StratisCliGenerationError,
    StratisCliEnvironmentError,
    StratisCliStratisdVersionError,
)
from ._connection import get_object
from ._constants import TOP_OBJECT, Id, IdType
from ._data import (
    MOFilesystem,
    MOPool,
    ObjectManager,
    Pool,
    Filesystem,
    filesystems,
    pools,
)

__all__ = [
    "check_stratisd_version",
    "StratisdErrors",
    "StratisCliGenerationError",
    "StratisCliEnvironmentError",
    "StratisCliStratisdVersionError",
    "get_object",
    "TOP_OBJECT",
    "Id",
    "IdType",
    "MOFilesystem",
    "MOPool",
    "ObjectManager",
    "Pool",
    "Filesystem",
    "filesystems",
    "pools",
]



# Copyright Red Hat
#
# snapm/manager/_loader.py - Snapshot Manager plugin loader
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: Apache-2.0
"""
Plugin loader logic for Snapshot Manager.
"""
import inspect
import importlib
import logging
from pathlib import Path
from importlib.util import find_spec
from snapm.manager.plugins import Plugin
import snapm.manager.plugins as plugin_pkg

_log = logging.getLogger(__name__)

_log_debug = _log.debug
_log_info = _log.info
_log_warn = _log.warning
_log_error = _log.error


def _find_plugin_modules(path: Path):
    for file in path.glob("[a-zA-Z]*.py"):
        if not file.name.startswith("_") and file.name != "__init__.py":
            yield file.stem


def _import_plugin_module(fqname: str):
    if not find_spec(fqname):
        return None
    try:
        _log_debug("Importing plugin module %s", fqname)
        return importlib.import_module(fqname)
    except (ModuleNotFoundError, ImportError) as e:
        _log_error("Error importing plugin %s: %s", fqname, e)
        return None


def _find_plugins_in_module(module, base_class):
    members = inspect.getmembers(module, inspect.isclass)
    names_to_check = getattr(module, "__all__", [name for name, _ in members])

    return [
        cls
        for name, cls in members
        if name in names_to_check
        and not name.startswith("_")
        and issubclass(cls, base_class)
        and cls is not base_class
        and cls.__module__ == module.__name__
    ]


def load_plugins(base_class=Plugin):
    """
    Attempt to load public plugins from ``plugin_pkg`` that are subclmsses of
    the class ``base_class``.
    """
    found_plugins = []

    for path in map(Path, plugin_pkg.__path__):
        for module_name in _find_plugin_modules(path):
            fqname = f"{plugin_pkg.__name__}.{module_name}"
            module = _import_plugin_module(fqname)
            if module:
                found_plugins.extend(_find_plugins_in_module(module, base_class))

    return sorted(found_plugins, key=lambda cls: cls.__name__)



# Copyright Red Hat
#
# snapm/manager/__init__.py - Snapshot Manager
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: Apache-2.0
"""
Top level interface to the snapshot manager.
"""

from ._manager import *  # noqa: F401, F403
from ._manager import __all__  # noqa: F401



# Copyright Red Hat
#
# snapm/report.py - Text reporting
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: Apache-2.0
"""The snapm reporting module contains a set of classes for creating
simple text based tabular reports for a user-defined set of object
types and fields. No restrictions are placed on the types of object
that can be reported: users of the ``Report`` classes may define
additional object types outside the ``snapm`` package and include these
types in reports generated by the module.

The fields displayed in a specific report may be selected from the
available set of fields by specifying a simple comma-separated string
list of field names (in display order). In addition, custom multi-column
sorting is possible using a similar string notation.

The ``Report`` module is closely based on the ``device-mapper``
reporting engine and shares many features and behaviours with device
mapper reports.
"""
import logging
import sys
import uuid
from json import dumps

from snapm import size_fmt


def find_minimum_sha_prefix(shas, min_prefix):
    """Find the minimum SHA prefix length guaranteeing uniqueness.

    Find the minimum unique prefix for the set of SHA IDs in the set
    ``shas``.

    :param shas: A set of SHA IDs
    :param min_prefix: Initial minimum prefix value
    :returns: The minimum unique prefix length for the set
    :rtype: int
    """
    shas = list(shas)
    shas.sort()
    for sha in shas:
        if shas.index(sha) == len(shas) - 1:
            continue

        def _next_sha(shas, sha):
            return shas[shas.index(sha) + 1]

        while sha[:min_prefix] == _next_sha(shas, sha)[:min_prefix]:
            min_prefix += 1
    return min_prefix


_log = logging.getLogger(__name__)

_log_debug = _log.debug
_log_info = _log.info
_log_warn = _log.warning
_log_error = _log.error

_DEFAULT_COLUMNS = 80

REP_NUM = "num"
REP_STR = "str"
REP_SHA = "sha"
REP_TIME = "time"
REP_UUID = "uuid"
REP_SIZE = "size"
REP_STR_LIST = "strlist"

_dtypes = [REP_NUM, REP_STR, REP_SHA, REP_TIME, REP_UUID, REP_SIZE, REP_STR_LIST]

_left_align_dtypes = [REP_STR, REP_SHA, REP_TIME, REP_UUID, REP_STR_LIST]
_right_align_dtypes = [REP_NUM, REP_SIZE]

_DEFAULT_WIDTH = 8

ALIGN_LEFT = "left"
ALIGN_RIGHT = "right"

_align_types = [ALIGN_LEFT, ALIGN_RIGHT]

ASCENDING = "ascending"
DESCENDING = "descending"

STANDARD_QUOTE = "'"
STANDARD_PAIR = "="

MIN_SHA_WIDTH = 7

string_types = (str,)
num_types = (int, float)


# pylint: disable=too-many-instance-attributes
class ReportOpts:
    """
    ReportOpts()
    Options controlling the formatting and output of a report.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        columns=_DEFAULT_COLUMNS,
        headings=True,
        buffered=True,
        separator=" ",
        field_name_prefix="",
        unquoted=True,
        aligned=True,
        json=False,
        columns_as_rows=False,
        report_file=sys.stdout,
    ):
        """Initialise ReportOpts object.

        Initialise a ``ReportOpts`` object to control output
        of a ``Report``.

        :param columns: the number of columns to use for output.
        :param headings: a boolean indicating whether to output
                         column headings for this report.
        :param buffered: a boolean indicating whether to buffer
                         output from this report.
        :param report_file: a file to which output will be sent.
        :returns: a new ``ReportOpts`` object.
        :rtype: ``<class ReportOpts>``
        """
        self.columns = columns
        self.headings = headings
        self.buffered = buffered
        self.separator = separator
        self.field_name_prefix = field_name_prefix
        self.unquoted = unquoted
        self.aligned = aligned
        self.json = json
        self.columns_as_rows = columns_as_rows
        self.report_file = report_file

    def __str__(self):
        return (
            f"columns={self.columns}\n"
            f"headings={self.headings}\n"
            f"buffered={self.buffered}\n"
            f"separator={self.separator}\n"
            f"field_name_prefix={self.field_name_prefix}\n"
            f"unquoted={self.unquoted}\n"
            f"aligned={self.aligned}\n"
            f"json={self.json}\n"
            f"columns_as_rows={self.columns_as_rows}\n"
            f"report_file={self.report_file}"
        )

    def __eq__(self, other):
        if not isinstance(other, ReportOpts):
            return False
        return all(
            [
                self.columns == other.columns,
                self.headings == other.headings,
                self.buffered == other.buffered,
                self.separator == other.separator,
                self.field_name_prefix == other.field_name_prefix,
                self.unquoted == other.unquoted,
                self.aligned == other.aligned,
                self.json == other.json,
                self.columns_as_rows == other.columns_as_rows,
                self.report_file == other.report_file,
            ]
        )


# pylint: disable=too-few-public-methods
class ReportObjType:
    """
    ReportObjType()

    Class representing a type of object to be reported on.
    Instances of ``ReportObjType`` must specify an identifier,
    a description, and a data function that will return the correct
    type of object from a compound object containing data objects
    of different types. For reports that use only a single object
    type the ``data_fn`` member may be simply ``lambda x: x``.
    """

    def __init__(self, objtype, desc, prefix, data_fn):
        """
        Initialise ReportObjType.

        Initialise a new ``ReportObjType`` object with the
        specified ``objtype``, ``desc``, optional ``prefix`` and
        ``data_fn``. The ``objtype`` must be an integer power of two
        that is unique within a given report. The ``data_fn`` should
        accept an object as its only argument and return an object
        of the requested type.
        """
        if not objtype or objtype < 0:
            raise ValueError("ReportObjType objtype cannot be <= 0.")

        if not desc:
            raise ValueError("ReportObjType desc cannot be empty.")

        if not data_fn:
            raise ValueError("ReportObjType requires data_fn.")

        self.objtype = objtype
        self.desc = desc
        self.prefix = prefix
        self.data_fn = data_fn


class FieldType:
    """
    FieldType()

    The ``FieldType`` class describes the properties of a field
    available in a ``Report`` instance.
    """

    # pylint: disable=too-many-arguments
    def __init__(self, objtype, name, head, desc, width, dtype, report_fn, align=None):
        """
        Initialise new FieldType object.

        Initialise a new ``FieldType`` object with the specified
        properties.

        :param objtype: The numeric object type ID (power of two)
        :param name: The field name used to select display fields
        :param desc: A human-readable description of the field
        :param width: The default (initial) field width
        :param dtype: The Report data type of the field
        :param report_fn: The field reporting function
        :param align: The field alignment value
        :returns: A new ReportFieldType object
        :rtype: ReportFieldType
        """
        if not objtype:
            raise ValueError("'objtype' must be non-zero")
        if not name:
            raise ValueError("'name' is required")
        self.objtype = objtype
        self.name = name
        self.head = head
        self.desc = desc

        if dtype not in _dtypes:
            raise ValueError(f"Invalid field dtype: {dtype}")

        if align and align not in _align_types:
            raise ValueError(f"Invalid field alignment: {align}")

        self.dtype = dtype
        self.report_fn = report_fn

        if not align:
            if dtype in _left_align_dtypes:
                self.align = ALIGN_LEFT
            elif dtype in _right_align_dtypes:
                self.align = ALIGN_RIGHT
        else:
            self.align = align

        if width < 0:
            raise ValueError("Field width cannot be < 0")
        self.width = width if width else _DEFAULT_WIDTH


class FieldProperties:
    """
    Properties of a field instance.
    """

    def __init__(
        self,
        field_num=None,
        initial_width=0,
        width=0,
        objtype=None,
        dtype=None,
        align=None,
        hidden=False,
        implicit=False,
        sort_key=False,
        sort_dir=None,
        compact_one=False,
        compacted=False,
        sort_posn=None,
    ):
        self.field_num = field_num
        self.initial_width = initial_width
        self.width = width
        self.objtype = objtype
        self.dtype = dtype
        self.align = align
        #
        # Field flags
        #
        self.hidden = hidden
        self.implicit = implicit
        self.sort_key = sort_key
        self.sort_dir = sort_dir
        self.compact_one = compact_one  # used for implicit fields
        self.compacted = compacted
        self.sort_posn = sort_posn


class Field:
    """
    Field()

    A ``Field`` represents an instance of a ``FieldType``
    including its associated data values.
    """

    #: reference to the containing Report
    def __init__(self, report, props):
        """
        Initialise a new Field object.

        Initialise a Field object and configure the supplied
        ``report`` and ``props`` attributes.

        :param report: The Report that owns this field
        :param props: The FieldProperties object for this field
        """
        self.report = report
        self.props = props
        self.report_string = ""
        self.sort_value = None

    def report_str(self, value):
        """
        Report a string value for this Field object.

        Set the value for this field to the supplied ``value``.

        :param value: The string value to set
        :rtype: None
        """
        if not isinstance(value, string_types):
            raise TypeError("Value for report_str() must be a string type.")
        self.set_value(value, sort_value=value)

    def report_sha(self, value):
        """
        Report a SHA value for this Field object.

        Set the value for this field to the supplied ``value``.

        :param value: The SHA value to set
        :rtype: None
        """
        if not isinstance(value, string_types):
            raise TypeError("Value for report_sha() must be a string type.")
        self.set_value(value, sort_value=value)

    def report_num(self, value):
        """
        Report a numeric value for this Field object.

        Set the value for this field to the supplied ``value``.

        :param value: The numeric value to set
        :rtype: None
        """
        if value is not None and not isinstance(value, num_types):
            raise TypeError("Value for report_num() must be a numeric type.")
        report_string = str(value)
        sort_value = value if value is not None else -1
        self.set_value(report_string, sort_value=sort_value)

    def report_time(self, value):
        """
        Report a time value for this Field object.

        Set the value for this field to the supplied ``value``.

        :param value: The time value to set
        :rtype: None
        """
        if not isinstance(value, string_types):
            raise TypeError("Value for report_time() must be a string type.")
        self.set_value(value, sort_value=value)

    def report_uuid(self, value):
        """
        Report a uuid value for this Field object.

        Set the value for this field to the supplied ``value``.

        :param value: The uuid value to set
        :rtype: None
        """
        if isinstance(value, uuid.UUID):
            value = str(value)
        else:
            raise TypeError("Value for report_uuid() must be a UUID type.")
        self.set_value(value, sort_value=value)

    def report_size(self, value):
        """
        Report a size value for this Field object.

        Set the value for this field to the supplied ``value``, converted
        to a human readable string.

        :param value: The size value to set in bytes.
        :rtype: None
        """
        if isinstance(value, int):
            value_str = size_fmt(value)
        else:
            raise TypeError("Value for report_size() must be an int type.")
        self.set_value(value_str, sort_value=value)

    def report_str_list(self, value):
        """
        Report a string list value for this Field object.

        Set the value for this field to the strings contained in ``value``.

        :param value: A list of strings
        :rtype: None
        """
        if not isinstance(value, list):
            raise TypeError("Value for report_str_list() must be a list type.")
        if not all(isinstance(v, string_types) for v in value):
            raise TypeError("Value for report_str_list() must be a list of strings.")
        value = sorted(value)
        list_value = ", ".join(value)
        self.set_value(list_value, sort_value=list_value)

    def set_value(self, report_string, sort_value=None):
        """
        Report an arbitrary value for this Field object.

        Set the value for this field to the supplied ``value``,
        and set the field's ``sort_value`` to the supplied
        ``sort_value``.

        :param report_string: The string value to set
        :param sort_value: The sort value
        :rtype: None
        """
        if report_string is None:
            raise ValueError("No value assigned to field.")
        self.report_string = report_string
        self.sort_value = sort_value if sort_value else report_string


class Row:
    """
    Row()

    A class representing a single data row making up a report.
    """

    def __init__(self, report):
        self.report = report
        self._fields = []
        self.sort_fields = None

    def add_field(self, field):
        """
        Add a field to this Row.

        :param field: The field to be added
        :rtype: None
        """
        self._fields.append(field)


def __none_returning_fn(_obj):
    """
    Dummy data function for special report types.

    :returns: None
    """
    return None


# Implicit report fields and types

PR_SPECIAL = 0x80000000
_implicit_special_report_types = [
    ReportObjType(PR_SPECIAL, "Special", "special_", __none_returning_fn)
]


def __no_report_fn(_field, _data):
    """
    Dummy report function for special report types.

    :returns: None
    """
    return


_SPECIAL_FIELD_HELP_NAME = "help"

_implicit_special_report_fields = [
    FieldType(
        PR_SPECIAL,
        _SPECIAL_FIELD_HELP_NAME,
        "Help",
        "Show help",
        8,
        REP_STR,
        __no_report_fn,
    )
]


# Report class


class Report:
    """
    Report()

    A class representing a configurable text report with multiple
    caller-defined fields. An optional title may be provided and the
    ``fields`` argument must contain a list of ``Field`` objects
    describing the required report.

    """

    # Implicit field support
    _implicit_types = _implicit_special_report_types
    _implicit_fields = _implicit_special_report_fields

    opts = None

    def __help_requested(self):
        """
        Check for presence of 'help' fields in output selection.

        Check the fields making up this Report and return True
        if any valid 'help' field synonym is present.

        :returns: True if help was requested or False otherwise
        """
        for field_prop in self._field_properties:
            if field_prop.implicit:
                name = self._implicit_fields[field_prop.field_num].name
                if name == _SPECIAL_FIELD_HELP_NAME:
                    return True
        return False

    def __get_longest_field_name_len(self, fields):
        """
        Find the longest field name length.

        :returns: the length of the longest configured field name
        """
        max_len = 0
        for field in fields:
            cur_len = len(field.name)
            max_len = cur_len if cur_len > max_len else max_len
        for obj_type in self._types:
            cur_len = len(obj_type.prefix) + 3
            max_len = cur_len if cur_len > max_len else max_len
        return max_len

    def __display_fields(self, display_field_types):
        """
        Display report fields help message.

        Display a list of valid fields for this ``Report``.

        :param fields: The list of fields to display
        :param display_field_types: A boolean controlling whether
                                    field types (str, SHA, num)
                                    are included in help output
        """
        out = self.opts.report_file
        fields = self._fields
        name_len = self.__get_longest_field_name_len(fields)
        last_desc = ""
        banner = "-" * 79
        for field in fields:
            obj_type = self.__find_type(field.objtype)
            if obj_type:
                desc = obj_type.desc
            else:
                desc = ""
            if desc != last_desc:
                if last_desc:
                    out.write(" \n")
                desc_len = len(desc) + 7
                out.write(f"{desc} Fields\n")
                out.write(f"{banner:{desc_len}.{desc_len}}\n")
            if display_field_types:
                open_type = " ["
                field_dtype = field.dtype
                close_type = "]"
            else:
                open_type = field_dtype = close_type = ""
            out.write(
                f"  {field.name:{name_len}} - {field.desc}"
                f"{open_type}{field_dtype}{close_type}\n"
            )
            last_desc = desc

    def __find_type(self, report_type):
        """
        Resolve numeric type to corresponding ReportObjType.

        :param report_type: The numeric report type to look up
        :returns: The requested ReportObjType.
        :raises: ValueError if no matching type was found.
        """
        for obj_type in self._implicit_types:
            if obj_type.objtype == report_type:
                return obj_type
        for obj_type in self._types:
            if obj_type.objtype == report_type:
                return obj_type

        raise ValueError(f"Unknown report object type: {report_type}")

    def __copy_field(self, field_num, implicit):
        """
        Copy field definition to FieldProperties

        Copy values from a FieldType to FieldProperties.

        :param field_num: The number of this field (fields order)
        :param implicit: True if this field is implicit, else False
        """
        return FieldProperties(
            field_num=field_num,
            width=self._fields[field_num].width,
            initial_width=self._fields[field_num].width,
            implicit=implicit,
            objtype=self.__find_type(self._fields[field_num].objtype),
            dtype=self._fields[field_num].dtype,
            align=self._fields[field_num].align,
        )

    def __add_field(self, field_num, implicit):
        """
        Add a field to this Report.

        Add the specified FieldType to this Report and configure
        FieldProperties for it.

        :param field_num: The number of this field (fields order)
        :param implicit: True if this field is implicit, else False
        """
        field_props = self.__copy_field(field_num, implicit)
        if field_props.hidden:
            self._field_properties.insert(0, field_props)
        else:
            self._field_properties.append(field_props)
        return field_props

    def __get_field(self, field_name):
        """
        Look up a field by name.

        Attempt to find the field named in ``field_name`` in this
        Report's tables of implicit and user-defined fields,
        returning a ``(field, implicit)`` tuple, where field
        contains the requested ``FieldType``, and ``implicit``
        is a boolean indicating whether this field is implicit or
        not.

        :param field_num: The number of this field (fields order)
        :param implicit: True if this field is implicit, else False
        """
        for field in self._implicit_fields:
            if field.name == field_name:
                return (self._implicit_fields.index(field), True)
        for field in self._fields:
            objtype = self.__find_type(field.objtype)
            if field_name in (field.name, objtype.prefix + field.name):
                return (self._fields.index(field), False)
        raise ValueError(f"No matching field name: {field_name}")

    def __field_match(self, field_name, type_only):
        """
        Attempt to match a field and optionally update report type.

        Look up the named field and, if ``type_only`` is True,
        update this Report's ``report_types`` mask to include
        the field's type identifier. If ``type_only`` is False the
        field is also added to this Report's field list.

        :param field_name: A string identifying the field
        :param type_only: True if this call should only update types
        """
        if "_" in field_name and field_name.split("_", maxsplit=1)[1] == "all":
            prefix = field_name.split("_", maxsplit=1)[0]
            for field in self._fields:
                objtype = self.__find_type(field.objtype)
                if objtype.prefix[:-1] == prefix:
                    self.__field_match(prefix + "_" + field.name, type_only)
            return None

        try:
            (f_idx, implicit) = self.__get_field(field_name)
            if type_only:
                if implicit:
                    self.report_types |= self._implicit_fields[f_idx].objtype
                else:
                    self.report_types |= self._fields[f_idx].objtype
                return None
            return self.__add_field(f_idx, implicit)
        except ValueError as err:
            _log_error("Error adding field %s", field_name)
            raise err
        return None

    def __parse_fields(self, field_format, type_only):
        """
        Parse report field list.

        Parse ``field_format`` and attempt to match the names of
        field names found to registered FieldType fields.

        If ``type_only`` is True only the ``report_types`` field
        is updated: otherwise the parsed fields are added to the
        Report's field list.

        :param field_format: The list of fields to parse
        :param type_only: True if this call should only update types
        """
        for word in field_format.split(","):
            # Allow consecutive commas
            if not word:
                continue
            try:
                self.__field_match(word, type_only)
            except ValueError as err:
                self.__display_fields(True)
                _log_error("Unrecognised field: %s", word)
                raise err

    def __add_sort_key(self, field_num, sort, implicit, type_only):
        """
        Add a new sort key to this Report

        Add the sort key identified by ``field_num`` to this list
        of sort keys for this Report.

        :param field_num: The field number of the key to add
        :param sort: The sort direction for this key
        :param implicit: True if field_num is implicit, else False
        :param type_only: True if this call should only update types
        """
        fields = self._implicit_fields if implicit else self._fields
        found = None

        for field_props in self._field_properties:
            if field_props.implicit == implicit and field_props.field_num == field_num:
                found = field_props

        if not found:
            if type_only:
                self.report_types |= fields[field_num].objtype
                return
            found = self.__add_field(field_num, implicit)

        if found.sort_key:
            _log_info("Ignoring duplicate sort field: %s", fields[field_num].name)
        found.sort_key = True
        found.sort_dir = sort
        found.sort_posn = self.keys_count
        self.keys_count += 1

    def __key_match(self, key_name, type_only):
        """
        Attempt to match a sort key and update report type.

        Look up the named sort key and, if ``type_only`` is True,
        update this Report's ``report_types`` mask to include
        the field's type identifier. If ``type_only`` is False the
        field is also added to this Report's field list.

        :param field_name: A string identifying the sort key
        :param type_only: True if this call should only update types
        """
        sort_dir = None

        if not key_name:
            raise ValueError("Sort key name cannot be empty")

        if key_name.startswith("+"):
            sort_dir = ASCENDING
            key_name = key_name[1:]
        elif key_name.startswith("-"):
            sort_dir = DESCENDING
            key_name = key_name[1:]
        else:
            sort_dir = ASCENDING

        for field in self._implicit_fields:
            fields = self._implicit_fields
            if field.name == key_name:
                return self.__add_sort_key(
                    fields.index(field), sort_dir, True, type_only
                )
        for field in self._fields:
            fields = self._fields
            objtype = self.__find_type(field.objtype)
            if key_name in (field.name, objtype.prefix + field.name):
                return self.__add_sort_key(
                    fields.index(field), sort_dir, False, type_only
                )

        raise ValueError(f"Unknown sort key name: {key_name}")

    def __parse_keys(self, keys, type_only):
        """
        Parse report sort key list.

        Parse ``keys`` and attempt to match the names of
        sort keys found to registered FieldType fields.

        If ``type_only`` is True only the ``report_types`` field
        is updated: otherwise the parsed fields are added to the
        Report's sort key list.

        :param field_format: The list of fields to parse
        :param type_only: True if this call should only update types
        """
        if not keys:
            return
        for word in keys.split(","):
            # Allow consecutive commas
            if not word:
                continue
            try:
                self.__key_match(word, type_only)
            except ValueError as err:
                self.__display_fields(True)
                _log_error("Unrecognised field: %s", word)
                raise err

    # pylint: disable=too-many-arguments
    def __init__(self, types, fields, output_fields, opts, sort_keys, title):
        """
        Initialise Report.

        Initialise a new ``Report`` object with the specified fields
        and output control options.

        :param types: List of ReportObjType used in this report.
        :param fields: A list of ``Field`` field descriptions.
        :param output_fields: An optional list of output fields to
                              be rendered by this report.
        :param opts: An instance of ``ReportOpts`` or None.
        :returns: A new report object.
        :rtype: ``Report``.
        """
        self.report_types = 0
        self.keys_count = 0

        self._data = None
        self._rows = None
        self._field_properties = None
        self._header_written = False
        self._field_calc_needed = True
        self._sort_required = False
        self._already_reported = False

        self._fields = fields
        self._types = types
        self._title = title

        if opts.buffered:
            self._sort_required = True

        self.opts = opts if opts else ReportOpts()

        self._rows = []
        self._field_properties = []

        # set field_prefix from type

        # canonicalize_field_ids()

        if not output_fields:
            output_fields = ",".join([field.name for field in fields])

        # First pass: set up types
        self.__parse_fields(output_fields, 1)
        self.__parse_keys(sort_keys, 1)

        # Second pass: initialise fields
        self.__parse_fields(output_fields, 0)
        self.__parse_keys(sort_keys, 0)

        if self.__help_requested():
            self._already_reported = True
            self.__display_fields(display_field_types=True)
            print("")

    def __recalculate_sha_width(self):
        """
        Recalculate minimum SHA field widths.

        For each REP_SHA field present, recalculate the minimum
        field width required to ensure uniqueness of the displayed
        values.

        :rtype: None
        """
        shas = {}
        props_map = {}
        for row in self._rows:
            for field in row._fields:
                if self._fields[field.props.field_num].dtype == REP_SHA:
                    # Use field_num as index to apply check across rows
                    num = field.props.field_num
                    if num not in shas:
                        shas[num] = set()
                        props_map[num] = field.props
                    shas[num].add(field.report_string)
        for num, vals in shas.items():
            min_prefix = max(MIN_SHA_WIDTH, props_map[num].width)
            props_map[num].width = find_minimum_sha_prefix(vals, min_prefix)

    def __recalculate_fields(self):
        """
        Recalculate field widths.

        For each field, recalculate the minimum field width by
        finding the longest ``report_string`` value for that field
        and updating the dynamic width stored in the corresponding
        ``FieldProperties`` object.

        :rtype: None
        """
        for row in self._rows:
            for field in row._fields:
                if self._sort_required and field.props.sort_key:
                    row.sort_fields[field.props.sort_posn] = field
                if self._fields[field.props.field_num].dtype == REP_SHA:
                    continue
                field_len = len(field.report_string)
                field.props.width = max(field_len, field.props.width)

    def __report_headings(self):
        """
        Output report headings.

        Output the column headings for this Report.

        :rtype: None
        """
        self._header_written = True
        if not self.opts.headings:
            return

        line = ""
        props = self._field_properties
        for field_props in props:
            if field_props.hidden:
                continue
            fields = self._fields
            heading = fields[field_props.field_num].head
            if self.opts.aligned:
                heading = f"{heading:{field_props.width}}"
            line += heading
            if props.index(field_props) != (len(props) - 1):
                line += self.opts.separator
        self.opts.report_file.write(line.strip() + "\n")

    def __row_key_fn(self):
        """
        Return a Python key function to compare report rows.

        The ``cmp`` argument of sorting functions has been removed
        in Python 3.x: to maintain similarity with the device-mapper
        report library we keep a traditional "cmp"-style function
        (that is structured identically to the version in the device
        mapper library), and dynamically wrap it in a ``RowKey``
        object to conform to the Python sort key model.

        :returns: A RowKey object wrapping _row_cmp()
        :rtype: RowKey
        """

        def _row_cmp(row_a, row_b):
            """
            Compare two report rows for sorting.

            Compare the report rows ``row_a`` and ``row_b`` and
            return a "cmp"-style comparison value:

                1 if row_a > row_b
                0 if row_a == row_b
               -1 if row_b < row_a

            Note that the actual comparison direction depends on the
            field definitions of the fields being compared, since
            each sort key defines its own sort order.

            :param row_a: The first row to compare
            :param row_b: The second row to compare
            """
            for cnt in range(0, row_a.report.keys_count):
                sfa = row_a.sort_fields[cnt]
                sfb = row_b.sort_fields[cnt]
                if sfa.props.dtype == REP_NUM:
                    num_a = sfa.sort_value
                    num_b = sfb.sort_value
                    if num_a == num_b:
                        continue
                    if sfa.props.sort_dir == ASCENDING:
                        return 1 if num_a > num_b else -1
                    return 1 if num_a < num_b else -1
                stra = sfa.sort_value
                strb = sfb.sort_value
                if stra == strb:
                    continue
                if sfa.props.sort_dir == ASCENDING:
                    return 1 if stra > strb else -1
                return 1 if stra < strb else -1
            return 0

        class RowKey:
            """
            RowKey sort wrapper.
            """

            def __init__(self, obj, *_args):
                """
                Initialise a new RowKey object.

                :param obj: The object to be compared
                :returns: None
                """
                self.obj = obj

            def __lt__(self, other):
                """
                Test if less than.

                :param other: The other object to be compared
                """
                return _row_cmp(self.obj, other.obj) < 0

            def __gt__(self, other):
                """
                Test if greater than.

                :param other: The other object to be compared
                """
                return _row_cmp(self.obj, other.obj) > 0

            def __eq__(self, other):
                """
                Test if equal to.

                :param other: The other object to be compared
                """
                return _row_cmp(self.obj, other.obj) == 0

            def __le__(self, other):
                """
                Test if less than or equal to.

                :param other: The other object to be compared
                """
                return _row_cmp(self.obj, other.obj) <= 0

            def __ge__(self, other):
                """
                Test if greater than or equal to.

                :param other: The other object to be compared
                """
                return _row_cmp(self.obj, other.obj) >= 0

            def __ne__(self, other):
                """
                Test if not equal to.

                :param other: The other object to be compared
                """
                return _row_cmp(self.obj, other.obj) != 0

        return RowKey

    def _sort_rows(self):
        """
        Sort the rows of this Report.

        Sort this report's rows, according to the configured sort
        keys.

        :returns: None
        """
        self._rows.sort(key=self.__row_key_fn())

    def report_object(self, obj):
        """
        Report data for object.

        Add a row of data to this ``Report``. The ``data``
        argument should be an object of the type understood by this
        report's fields. It will be passed in turn to each field to
        obtain data for the current row.

        :param obj: the object to report on for this row.
        """
        if obj is None:
            raise ValueError("Cannot report NoneType object.")

        if self._already_reported:
            return ""

        row = Row(self)
        fields = self._fields
        if self._sort_required:
            row.sort_fields = [-1] * self.keys_count
        for field_props in self._field_properties:
            field = Field(self, field_props)
            data = field_props.objtype.data_fn(obj)

            if data is None:
                raise ValueError(
                    f"No data assigned to field {fields[field_props.field_num].name}"
                )

            try:
                fields[field_props.field_num].report_fn(field, data)
            except ValueError as err:
                raise ValueError(
                    f"No value assigned to field {fields[field_props.field_num].name}"
                ) from err
            row.add_field(field)
        self._rows.append(row)

        if not self.opts.buffered:
            return self.report_output()
        return ""

    def _output_field(self, field):
        """
        Output field data.

        Generate string data for one field in a report row.

        :field: The field to be output
        :returns: The output report string for this field
        :rtype: str
        """
        fields = self._fields
        prefix = self.opts.field_name_prefix
        quote = "" if self.opts.unquoted else STANDARD_QUOTE

        if prefix:
            field_name = fields[field.props.field_num].name
            prefix += f"{field_name.upper()}{STANDARD_PAIR}{STANDARD_QUOTE}"

        repstr = field.report_string
        width = field.props.width
        if self.opts.aligned:
            align = field.props.align
            if not align:
                if field.props.dtype in _right_align_dtypes:
                    align = ALIGN_RIGHT
                else:
                    align = ALIGN_LEFT
            if align == ALIGN_LEFT:
                repstr = f"{repstr: <{width}.{width}}"
            else:
                repstr = f"{repstr: >{width}.{width}}"

        suffix = quote
        return prefix + repstr + suffix

    def _output_field_json(self, field):
        """
        Output field data as a JSON dictionary element.

        :field: The field to be output
        :returns: A ``(key, value)`` tuple for this field
        :rtype: Tuple
        """
        fields = self._fields
        obj_type = field.props.objtype
        field_name = obj_type.prefix + fields[field.props.field_num].name
        field_dtype = fields[field.props.field_num].dtype
        if field_dtype == REP_NUM:
            value = field.sort_value
        elif field_dtype == REP_STR:
            if field.report_string == "yes":
                value = True
            elif field.report_string == "no":
                value = False
            else:
                value = field.report_string
        elif field_dtype == REP_STR_LIST:
            value = field.report_string.split(", ")
        else:
            value = field.report_string

        return (field_name, value)

    def _output_as_rows(self):
        """
        Output this report in column format.

        Output the data contained in this ``Report`` in column
        format, one row per line. If column headings have not been
        printed already they will be automatically displayed by this
        call.

        :returns: None
        """
        for field_props in self._field_properties:
            if field_props.hidden:
                for row in self._rows:
                    row._fields = row._fields[1:]

            fields = self._implicit_fields if field_props.implicit else self._fields
            line = ""

            if self.opts.headings:
                line += fields[field_props.field_num].head + self.opts.separator

            for row in self._rows:
                field = row._fields[0]
                line += self._output_field(field)
                line += self.opts.separator
                row._fields = row._fields[1:]

            self.opts.report_file.write(line.strip() + "\n")

    def _output_as_columns(self):
        """
        Output this report in column format.

        Output the data contained in this ``Report`` in column
        format, one row per line. If column headings have not been
        printed already they will be automatically displayed by this
        call.

        :returns: None
        """
        if not self._header_written:
            self.__report_headings()
        for row in self._rows:
            do_field_delim = False
            line = ""
            for field in row._fields:
                if field.props.hidden:
                    continue
                if do_field_delim:
                    line += self.opts.separator
                else:
                    do_field_delim = True
                line += self._output_field(field)
            self.opts.report_file.write(line.strip() + "\n")

    def _output_as_json(self):
        """
        Output this report in JSON format.

        Output the data contained in this ``Report`` in JSON notation.
        Column names are used as the keys for the JSON report.
        :returns: None
        """
        rows = {f"{self._title}": []}
        for row in self._rows:
            row_vals = {}
            for field in row._fields:
                if field.props.hidden:
                    continue
                (key, value) = self._output_field_json(field)
                row_vals[key] = value
            rows[f"{self._title}"].append(row_vals)
        self.opts.report_file.write(dumps(rows, indent=4) + "\n")

    def report_output(self):
        """
        Output report data.

        Output this report's data to the configured report file,
        using the configured output controls and fields.

        On success the number of rows output is returned. On
        error an exception is raised.

        :returns: the number of rows of output written.
        :rtype: ``int``
        """
        if self._already_reported:
            return ""
        if self._field_calc_needed:
            self.__recalculate_sha_width()
            self.__recalculate_fields()
        if self._sort_required:
            self._sort_rows()
        if self.opts.json:
            return self._output_as_json()
        if self.opts.columns_as_rows:
            return self._output_as_rows()
        return self._output_as_columns()


__all__ = [
    # Module constants
    "REP_NUM",
    "REP_STR",
    "REP_SHA",
    "REP_TIME",
    "REP_UUID",
    "REP_SIZE",
    "REP_STR_LIST",
    "ALIGN_LEFT",
    "ALIGN_RIGHT",
    "ASCENDING",
    "DESCENDING",
    # Report objects
    "ReportOpts",
    "ReportObjType",
    "Field",
    "FieldType",
    "FieldProperties",
    "Report",
]

# vim: set et ts=4 sw=4 :



# Copyright Red Hat
#
# tests/test_plugin.py - Plugin core unit tests
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: Apache-2.0
import unittest
import logging

log = logging.getLogger()

import snapm.manager.plugins as plugins


def _find_device_mounts():
    """
    Find mount points on the running system backed by a device in /dev.

    :returns: A list of device-backed mount points.
    """
    mount_points = []
    with open("/proc/mounts", "r", encoding="utf8") as mounts:
        for line in mounts.readlines():
            (device, mount_point, _) = line.split(maxsplit=2)
            if device.startswith("/dev"):
                mount_points.append(mount_point)
    return mount_points


class PluginTests(unittest.TestCase):
    """Test plugin helper functions"""

    def setUp(self):
        log.debug("Preparing %s", self._testMethodName)

    def tearDown(self):
        log.debug("Tearing down %s", self._testMethodName)

    def test_plugin_info(self):
        p = plugins.Plugin(log)
        info = p.info()
        self.assertEqual(info, {"name": "plugin", "version": "0.1.0"})

    def _plugin_base_not_implemented_raises(self, method, args):
        p = plugins.Plugin(log)
        with self.assertRaises(NotImplementedError):
            getattr(p, method)(*args)

    def test_plugin_base_discover_snapshots_raises(self):
        args = []
        self._plugin_base_not_implemented_raises("discover_snapshots", args)

    def test_plugin_base_can_snapshot_raises(self):
        args = ["/mount"]
        self._plugin_base_not_implemented_raises("can_snapshot", args)

    def test_plugin_base_check_create_snapshot_raises(self):
        args = [
            "fedora/root",
            "testset0",
            1693921253,
            "/",
            None,
        ]
        self._plugin_base_not_implemented_raises("check_create_snapshot", args)

    def test_plugin_base_create_snapshot_raises(self):
        args = [
            "fedora/root",
            "testset0",
            1693921253,
            "/",
            None,
        ]
        self._plugin_base_not_implemented_raises("create_snapshot", args)

    def test_plugin_base_rename_snapshot_raises(self):
        args = [
            "old_name",
            "fedora/root",
            "testset0",
            1693921253,
            "/",
        ]
        self._plugin_base_not_implemented_raises("rename_snapshot", args)

    def test_plugin_base_delete_snapshot_raises(self):
        args = ["aname"]
        self._plugin_base_not_implemented_raises("delete_snapshot", args)

    def test_plugin_base_activate_snapshot_raises(self):
        args = ["aname"]
        self._plugin_base_not_implemented_raises("activate_snapshot", args)

    def test_plugin_base_deactivate_snapshot_raises(self):
        args = ["aname"]
        self._plugin_base_not_implemented_raises("deactivate_snapshot", args)

    def test_plugin_base_set_autoactivate_raises(self):
        args = ["aname"]
        self._plugin_base_not_implemented_raises("set_autoactivate", args)

    def test_plugin_base_origin_from_mount_point_raises(self):
        args = ["/mount"]
        self._plugin_base_not_implemented_raises("origin_from_mount_point", args)

    def test_plugin_base_check_resize_snapshot_raises(self):
        args = [
            "aname",
            "/dev/fedora/aname-orig",
            "/mount",
            "1G",
        ]
        self._plugin_base_not_implemented_raises("check_resize_snapshot", args)

    def test_plugin_base_resize_snapshot_raises(self):
        args = [
            "aname",
            "/dev/fedora/aname-orig",
            "/mount",
            "1G",
        ]
        self._plugin_base_not_implemented_raises("resize_snapshot", args)

    def test_plugin_base_check_revert_snapshot_raises(self):
        args = [
            "aname",
            "/dev/fedora/aname-orig",
        ]
        self._plugin_base_not_implemented_raises("check_revert_snapshot", args)

    def test_plugin_base_revert_snapshot_raises(self):
        args = ["aname"]
        self._plugin_base_not_implemented_raises("revert_snapshot", args)

    def test_parse_snapshot_name(self):
        snapshot_names = {
            "root-snapset_backup_1693921253_-": ("backup", 1693921253, "/"),
            "data-snapset_backup_1693921253_-data": (
                "backup",
                1693921253,
                "/data",
            ),
            "home-snapset_myset_1693921266_-home": (
                "myset",
                1693921266,
                "/home",
            ),
            "srv-snapset_newset_1694704292_-srv": ("newset", 1694704292, "/srv"),
        }
        for snapshot_name in snapshot_names.keys():
            (origin, _) = snapshot_name.split("-", maxsplit=1)
            self.assertEqual(
                plugins.parse_snapshot_name(snapshot_name, origin),
                snapshot_names[snapshot_name],
            )

    def test_parse_snapshot_name_none(self):
        snapshot_names = {
            "some_snapshot_volume": "origin_volume",
            "somesnapshot": "some",
            "some_snapshot": "some",
            "root-notsnapset_backup_1693921253_-": "root",
        }
        for name, origin in snapshot_names.items():
            self.assertEqual(None, plugins.parse_snapshot_name(name, origin))

    def test_device_from_mount_point(self):
        mounts = _find_device_mounts()
        for mount in mounts:
            dev = plugins.device_from_mount_point(mount)
            self.assertTrue(dev.startswith("/dev"))

    def test_device_from_mount_point_bad_mount(self):
        with self.assertRaises(KeyError) as cm:
            dev = plugins.device_from_mount_point("/quuxfoo")

    def test_mount_point_space_used(self):
        used = plugins.mount_point_space_used("/sys")
        self.assertEqual(0, used)

    def test_encode_mount_point(self):
        mounts = {
            "/": "-",
            "/home": "-home",
            "/var": "-var",
            "/opt": "-opt",
            "/var/tmp": "-var-tmp",
            "/var/foo-bar": "-var-foo--bar",
            "/foo_bar": "-foo_bar",
            "/data:storage": "-data.3astorage",
            "/data.storage": "-data..storage",
        }
        for mount in mounts.keys():
            self.assertEqual(plugins.encode_mount_point(mount), mounts[mount])

    def test_decode_mount_point(self):
        enc_mounts = {
            "-": "/",
            "-home": "/home",
            "-var": "/var",
            "-opt": "/opt",
            "-var-tmp": "/var/tmp",
            "-var-foo--bar": "/var/foo-bar",
            "-foo_bar": "/foo_bar",
            "-data.3astorage": "/data:storage",
            "-data..storage" : "/data.storage",
        }
        for enc_mount in enc_mounts.keys():
            self.assertEqual(
                plugins.decode_mount_point(enc_mount), enc_mounts[enc_mount]
            )

    def test_format_snapshot_name(self):
        snapshot_parts = {
            (
                "data",
                "backup",
                1693921253,
                "/data",
            ): "data-snapset_backup_1693921253_-data",
            (
                "home",
                "myset",
                1693921266,
                "/home",
            ): "home-snapset_myset_1693921266_-home",
            (
                "srv",
                "newset",
                1694704292,
                "/srv",
            ): "srv-snapset_newset_1694704292_-srv",
        }
        for parts in snapshot_parts.keys():
            mount = plugins.encode_mount_point(parts[3])
            self.assertEqual(
                plugins.format_snapshot_name(parts[0], parts[1], parts[2], mount),
                snapshot_parts[parts],
            )

    def test__parse_proc_mounts_line(self):
        from snapm.manager.plugins._plugin import _parse_proc_mounts_line

        lines = {
            "/dev/mapper/fedora-root / xfs rw": (
                "/dev/mapper/fedora-root",
                "/",
            ),
            "/dev/mapper/fedora-data /data xfs rw": (
                "/dev/mapper/fedora-data",
                "/data",
            ),
            "/dev/vda2 /boot xfs rw": ("/dev/vda2", "/boot"),
        }
        for line in lines:
            self.assertEqual(_parse_proc_mounts_line(line), lines[line])



# Copyright Red Hat
#
# tests/test_manager.py - Manager core unit tests
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: Apache-2.0
import unittest
import logging
import os
from uuid import UUID


import snapm
import snapm.manager as manager
import boom

from tests import have_root, BOOT_ROOT_TEST
from ._util import LvmLoopBacked, StratisLoopBacked


log = logging.getLogger()

boom.set_boot_path(BOOT_ROOT_TEST)


class ManagerTestsSimple(unittest.TestCase):
    """
    Test manager interfaces with mock callouts
    """

    _old_path = None

    def setUp(self):
        log.debug("Preparing %s", self._testMethodName)
        bin_path = os.path.abspath("tests/bin")
        cur_path = os.environ["PATH"]
        self._old_path = cur_path
        os.environ["PATH"] = bin_path + os.pathsep + cur_path

    def tearDown(self):
        log.debug("Tearing down %s", self._testMethodName)
        os.environ["PATH"] = self._old_path

    def test_manager(self):
        m = manager.Manager()
        self.assertEqual(len(m.snapshot_sets), 3)

    def test_find_snapshot_sets_all(self):
        m = manager.Manager()
        s = snapm.Selection()
        sets = m.find_snapshot_sets(selection=s)
        self.assertEqual(len(sets), 3)

    def test_find_snapshots_all(self):
        m = manager.Manager()
        s = snapm.Selection()
        snaps = m.find_snapshots(selection=s)
        self.assertEqual(len(snaps), 16)

    def test_snapset_to_str(self):
        m = manager.Manager()
        sets = m.find_snapshot_sets()
        set_str = str(sets[0])
        self.assertTrue(isinstance(set_str, str))

    def test_snapshot_to_str(self):
        m = manager.Manager()
        snaps = m.find_snapshots()
        snap_str = str(snaps[0])
        self.assertTrue(isinstance(snap_str, str))

    def test_snapset_time(self):
        m = manager.Manager()
        s = snapm.Selection(name="backup")
        sets = m.find_snapshot_sets(selection=s)
        self.assertEqual(sets[0].time, "2023-09-05 13:40:53")

    def test_snapshot_time(self):
        m = manager.Manager()
        s = snapm.Selection(name="backup")
        sets = m.find_snapshot_sets(selection=s)
        snap = sets[0].snapshots[0]
        self.assertEqual(snap.time, "2023-09-05 13:40:53")

    def test_snapset_mount_points(self):
        m = manager.Manager()
        s = snapm.Selection(name="backup")
        sets = m.find_snapshot_sets(selection=s)
        self.assertEqual(sets[0].mount_points, ["/home", "/opt", "/", "/data", "/srv"])

    def test_snapset_find_by_mount_points(self):
        m = manager.Manager()
        s = snapm.Selection(mount_points=["/home", "/opt", "/", "/data", "/srv"])
        sets = m.find_snapshot_sets(selection=s)
        self.assertEqual(len(sets), 2)

    def test_snapset_find_by_timestamp(self):
        m = manager.Manager()
        s = snapm.Selection(timestamp=1693921253)
        sets = m.find_snapshot_sets(selection=s)
        self.assertEqual(len(sets), 1)

    def test_snapset_find_by_nr_snapshots(self):
        m = manager.Manager()
        s = snapm.Selection(nr_snapshots=5)
        sets = m.find_snapshot_sets(selection=s)
        self.assertEqual(len(sets), 2)


@unittest.skipIf(not have_root(), "requires root privileges")
class ManagerTests(unittest.TestCase):
    """
    Tests for snapm.manager.Manager that apply to all supported snapshot
    providers.
    """
    volumes = ["root", "home", "var", "data_vol"]
    thin_volumes = ["opt", "srv", "thin-vol"]
    stratis_volumes = ["fs1", "fs2"]

    def setUp(self):
        log.debug("Preparing %s", self._testMethodName)
        def cleanup_lvm():
            log.debug("Cleaning up LVM (%s)", self._testMethodName)
            if hasattr(self, "_lvm"):
                self._lvm.destroy()

        def cleanup_stratis():
            log.debug("Cleaning up Stratis (%s)", self._testMethodName)
            if hasattr(self, "_stratis"):
                self._stratis.destroy()

        self.addCleanup(cleanup_lvm)
        self.addCleanup(cleanup_stratis)

        self._lvm = LvmLoopBacked(self.volumes, thin_volumes=self.thin_volumes)
        self._stratis = StratisLoopBacked(self.stratis_volumes)

        self.manager = snapm.manager.Manager()

    def mount_points(self):
        return self._lvm.mount_points() + self._stratis.mount_points()

    def stop_start_storage(self):
        self._lvm.umount_all()
        self._lvm.deactivate()
        self._lvm.activate()
        self._lvm.mount_all()

        self._stratis.umount_all()
        self._stratis.stop_pool()
        self._stratis.start_pool()
        self._stratis.mount_all()

    def test_find_snapshot_sets_none(self):
        sets = self.manager.find_snapshot_sets()
        self.assertEqual([], sets)

    def test_find_snapshot_sets_one(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        sets = self.manager.find_snapshot_sets()
        self.assertEqual(len(sets), 1)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_find_snapshot_sets_with_selection_name(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        self.manager.create_snapshot_set("testset1", self.mount_points())
        s = snapm.Selection(name="testset0")
        sets = self.manager.find_snapshot_sets(selection=s)
        self.assertEqual(len(sets), 1)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset1"))

    def test_find_snapshot_sets_with_selection_uuid(self):
        set1 = self.manager.create_snapshot_set("testset0", self.mount_points())
        self.manager.create_snapshot_set("testset1", self.mount_points())
        s = snapm.Selection(uuid=set1.uuid)
        sets = self.manager.find_snapshot_sets(selection=s)
        self.assertEqual(len(sets), 1)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset1"))

    def test_create_snapshot_set(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        s = snapm.Selection(name="testset0")
        sets = self.manager.find_snapshot_sets(selection=s)
        self.assertEqual(len(sets), 1)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_create_snapshot_set_blockdevs(self):
        snapset = self.manager.create_snapshot_set(
            "testset0", self._lvm.block_devs() + self._stratis.block_devs()
        )
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_create_snapshot_set_blockdevs_unmounted(self):
        self._lvm.umount_all()
        self._stratis.umount_all()
        snapset = self.manager.create_snapshot_set(
            "testset0", self._lvm.block_devs() + self._stratis.block_devs()
        )
        self._lvm.mount_all()
        self._stratis.mount_all()
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_create_snapshot_set_mixed_1(self):
        self._stratis.umount_all()
        snapset = self.manager.create_snapshot_set(
            "testset0", self._lvm.mount_points() + self._stratis.block_devs()
        )
        self._stratis.mount_all()
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_create_snapshot_set_mixed_2(self):
        self._lvm.umount_all()
        snapset = self.manager.create_snapshot_set(
            "testset0", self._lvm.block_devs() + self._stratis.mount_points()
        )
        self._lvm.mount_all()
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_create_snapshot_set_default_size_policy(self):
        self.manager.create_snapshot_set(
            "testset0", self.mount_points(), default_size_policy="10%FREE"
        )
        s = snapm.Selection(name="testset0")
        sets = self.manager.find_snapshot_sets(selection=s)
        self.assertEqual(len(sets), 1)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_create_snapshot_set_size_policies_10_free(self):
        mount_specs = [f"{mp}:10%FREE" for mp in self.mount_points()]
        self.manager.create_snapshot_set("testset0", mount_specs)
        s = snapm.Selection(name="testset0")
        sets = self.manager.find_snapshot_sets(selection=s)
        self.assertEqual(len(sets), 1)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_create_snapshot_set_size_policies_10_size(self):
        mount_specs = [f"{mp}:10%SIZE" for mp in self.mount_points()]
        self.manager.create_snapshot_set("testset0", mount_specs)
        s = snapm.Selection(name="testset0")
        sets = self.manager.find_snapshot_sets(selection=s)
        self.assertEqual(len(sets), 1)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_create_snapshot_set_size_policies_200_used(self):
        mount_specs = [f"{mp}:200%USED" for mp in self.mount_points()]
        self.manager.create_snapshot_set("testset0", mount_specs)
        s = snapm.Selection(name="testset0")
        sets = self.manager.find_snapshot_sets(selection=s)
        self.assertEqual(len(sets), 1)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_create_snapshot_set_size_policies_100_size(self):
        mount_specs = [f"{mp}:100%SIZE" for mp in self.mount_points()]
        self.manager.create_snapshot_set("testset0", mount_specs)
        s = snapm.Selection(name="testset0")
        sets = self.manager.find_snapshot_sets(selection=s)
        self.assertEqual(len(sets), 1)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_create_snapshot_set_size_policies_fixed(self):
        mount_specs = [f"{mp}:512MiB" for mp in self.mount_points()]
        self.manager.create_snapshot_set("testset0", mount_specs)
        s = snapm.Selection(name="testset0")
        sets = self.manager.find_snapshot_sets(selection=s)
        self.assertEqual(len(sets), 1)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_create_snapshot_set_size_policy_size_over_100_raises(self):
        with self.assertRaises(snapm.SnapmSizePolicyError) as cm:
            self.manager.create_snapshot_set(
                "testset0", self.mount_points(), default_size_policy="150%SIZE"
            )

    def test_create_snapshot_set_size_policy_free_over_100_raises(self):
        with self.assertRaises(snapm.SnapmSizePolicyError) as cm:
            self.manager.create_snapshot_set(
                "testset0", self.mount_points(), default_size_policy="150%FREE"
            )

    def test_create_snapshot_set_size_policy_bad_units_raises(self):
        with self.assertRaises(snapm.SnapmSizePolicyError) as cm:
            self.manager.create_snapshot_set(
                "testset0", self.mount_points(), default_size_policy="2FiB"
            )

    def test_create_snapshot_set_size_policy_non_num_raises(self):
        with self.assertRaises(snapm.SnapmSizePolicyError) as cm:
            self.manager.create_snapshot_set(
                "testset0", self.mount_points(), default_size_policy="quux"
            )

    def test_create_snapshot_set_size_policies_blockdev_used_raises(self):
        dev_specs = [f"{dev}:100%USED" for dev in self._lvm.block_devs()]
        self._lvm.umount_all()
        with self.assertRaises(snapm.SnapmSizePolicyError) as cm:
            self.manager.create_snapshot_set("testset0", dev_specs)
        self._lvm.mount_all()

    def test_create_snapshot_set_bad_name_backslash(self):
        with self.assertRaises(snapm.SnapmInvalidIdentifierError) as cm:
            self.manager.create_snapshot_set("bad\\name", self.mount_points())

    def test_create_snapshot_set_bad_name_underscore(self):
        with self.assertRaises(snapm.SnapmInvalidIdentifierError) as cm:
            self.manager.create_snapshot_set("bad_name", self.mount_points())

    def test_create_snapshot_set_bad_name_slash(self):
        with self.assertRaises(snapm.SnapmInvalidIdentifierError) as cm:
            self.manager.create_snapshot_set("bad/name", self.mount_points())

    def test_create_snapshot_set_bad_name_space(self):
        with self.assertRaises(snapm.SnapmInvalidIdentifierError) as cm:
            self.manager.create_snapshot_set("bad name", self.mount_points())

    def test_create_snapshot_set_bad_name_at(self):
        with self.assertRaises(snapm.SnapmInvalidIdentifierError) as cm:
            self.manager.create_snapshot_set("bad@name", self.mount_points())

    def test_create_snapshot_set_bad_name_pipe(self):
        with self.assertRaises(snapm.SnapmInvalidIdentifierError) as cm:
            self.manager.create_snapshot_set("bad|name", self.mount_points())

    def test_create_snapshot_set_name_too_long(self):
        name = "a" * 127
        with self.assertRaises(snapm.SnapmInvalidIdentifierError) as cm:
            self.manager.create_snapshot_set(name, self.mount_points())

    def test_create_snapshot_set_no_space_raises(self):
        with self.assertRaises(snapm.SnapmNoSpaceError) as cm:
            for i in range(0, 10):
                self.manager.create_snapshot_set(
                    f"testset{i}", self.mount_points()
                )
                self._lvm.dump_lvs()

    def test_create_delete_snapshot_set(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        s = snapm.Selection(name="testset0")
        self.manager.delete_snapshot_sets(s)
        sets = self.manager.find_snapshot_sets(selection=s)
        self.assertEqual(len(sets), 0)

    def test_delete_snapshot_set_nosuch(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        s = snapm.Selection(name="nosuch")
        with self.assertRaises(snapm.SnapmNotFoundError) as cm:
            self.manager.delete_snapshot_sets(s)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_delete_snapshot_set_mounted(self):
        sset = self.manager.create_snapshot_set("testset0", self.mount_points())
        s = snapm.Selection(name="testset0")

        mnt_snap = sset.snapshots[0]
        mnt_name = mnt_snap.name.split("/")[1]

        self._lvm.make_mount_point(mnt_name)
        self._lvm.mount(mnt_name)

        with self.assertRaises(snapm.SnapmBusyError) as cm:
            self.manager.delete_snapshot_sets(s)

        self._lvm.umount(mnt_name)
        self.manager.delete_snapshot_sets(s)

    def test_delete_err_raises(self):
        sset = self.manager.create_snapshot_set("testset0", self.mount_points())
        selection = snapm.Selection(name="testset0")

        def fail_delete():
            raise snapm.SnapmError("Error deleting snapshot")

        orig_delete = sset.snapshots[1].delete
        sset.snapshots[1].delete = fail_delete

        with self.assertRaises(snapm.SnapmPluginError) as cm:
            self.manager.delete_snapshot_sets(selection)

        sset.snapshots[0].delete = orig_delete
        self.manager.delete_snapshot_sets(selection)

    def test_create_snapshot_set_duplicate(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        with self.assertRaises(snapm.SnapmExistsError) as cm:
            self.manager.create_snapshot_set("testset0", self.mount_points())
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_create_snapshot_set_not_a_mount_point(self):
        mount_point = self.mount_points()[0]
        non_mount = os.path.join(mount_point, "etc")
        os.makedirs(non_mount)
        with self.assertRaises(snapm.SnapmPathError) as cm:
            self.manager.create_snapshot_set("testset0", [non_mount])

    def test_create_snapshot_set_no_provider(self):
        with self.assertRaises(snapm.SnapmError) as cm:
            self.manager.create_snapshot_set("testset0", ["/boot"])

    def test_rename_snapshot_set(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        s = snapm.Selection(name="testset0")
        sets = self.manager.find_snapshot_sets(selection=s)
        self.assertEqual(len(sets), 1)
        self.manager.rename_snapshot_set("testset0", "testset1")
        s1 = snapm.Selection(name="testset1")
        sets = self.manager.find_snapshot_sets(selection=s1)
        self.assertEqual(len(sets), 1)
        sets = self.manager.find_snapshot_sets(selection=s)
        self.assertEqual(len(sets), 0)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset1"))

    def test_rename_snapshot_set_nosuch(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        with self.assertRaises(snapm.SnapmNotFoundError) as cm:
            self.manager.rename_snapshot_set("nosuch", "newname")
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_rename_snapshot_set_exists(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        self.manager.create_snapshot_set("testset1", self.mount_points())
        with self.assertRaises(snapm.SnapmExistsError) as cm:
            self.manager.rename_snapshot_set("testset0", "testset1")
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset1"))

    def test_rename_err_and_rollback(self):
        sset = self.manager.create_snapshot_set("testset0", self.mount_points())

        def fail_rename(_new_name):
            raise snapm.SnapmError("Error renaming snapshot")

        sset.snapshots[1].rename = fail_rename

        with self.assertRaises(snapm.SnapmPluginError) as cm:
            self.manager.rename_snapshot_set("testset0", "testset1")

        self.assertEqual(sset.name, "testset0")
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_find_snapshots(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        snaps = self.manager.find_snapshots()
        self.assertEqual(len(snaps), len(self.mount_points()))
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_find_snapshots_with_selection(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        self.manager.create_snapshot_set("testset1", self.mount_points())
        s = snapm.Selection(name="testset0")
        snaps = self.manager.find_snapshots(selection=s)
        self.assertEqual(len(snaps), len(self.mount_points()))
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset1"))

    def test_activate_deactivate_snapsets(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        s = snapm.Selection(name="testset0")
        self.manager.activate_snapshot_sets(selection=s)
        sets = self.manager.find_snapshot_sets(selection=s)
        for snap in sets[0].snapshots:
            self.assertEqual(snap.status, snapm.SnapStatus.ACTIVE)
        self.manager.deactivate_snapshot_sets(selection=s)
        for snap in sets[0].snapshots:
            if snap.origin.removeprefix("test_vg0/") in self.thin_volumes:
                self.assertEqual(snap.status, snapm.SnapStatus.INACTIVE)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_activate_snapsets_nosuch(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        s = snapm.Selection(name="nosuch")
        with self.assertRaises(snapm.SnapmNotFoundError) as cm:
            self.manager.activate_snapshot_sets(selection=s)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_deactivate_snapsets_nosuch(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        s = snapm.Selection(name="nosuch")
        with self.assertRaises(snapm.SnapmNotFoundError) as cm:
            self.manager.deactivate_snapshot_sets(selection=s)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_set_autoactivate_snapsets(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        s = snapm.Selection(name="testset0")
        self.manager.set_autoactivate(s, auto=False)
        sets = self.manager.find_snapshot_sets(selection=s)
        # autoactivate cannot be set to False for Stratis snapshots
        #for snap in sets[0].snapshots:
        #    self.assertEqual(snap.autoactivate, False)
        self.manager.set_autoactivate(s, auto=True)
        for snap in sets[0].snapshots:
            self.assertEqual(snap.autoactivate, True)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_set_autoactivate_snapsets_nosuch(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())
        s = snapm.Selection(name="nosuch")
        with self.assertRaises(snapm.SnapmNotFoundError) as cm:
            self.manager.set_autoactivate(s, True)
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_revert_snapshot_sets(self):
        # Revert is only supported on LVM2: don't include Stratis in the test.
        origin_file = "root/origin"
        snapshot_file = "root/snapshot"
        testset = "testset0"

        # Create files in the origin volume and post-snapshot
        self._lvm.touch_path(origin_file)
        self.manager.create_snapshot_set(testset, self.mount_points())
        self._lvm.touch_path(snapshot_file)

        # Test that the origin and snapshot files both exist
        self.assertEqual(self._lvm.test_path(origin_file), True)
        self.assertEqual(self._lvm.test_path(snapshot_file), True)

        # Start revert the snapshot set
        selection = snapm.Selection(name=testset)
        self.manager.revert_snapshot_sets(selection)

        # Unmount the set, deactivate/reactivate and re-mount
        self.stop_start_storage()

        # Test that only the origin file exists
        self.assertEqual(self._lvm.test_path(origin_file), True)
        self.assertEqual(self._lvm.test_path(snapshot_file), False)

    def test_revert_reverting_snapshot_set_raises(self):
        testset = "testset0"
        self.manager.create_snapshot_set(testset, self.mount_points())

        selection = snapm.Selection(name=testset)
        self.manager.revert_snapshot_sets(selection)

        with self.assertRaises(snapm.SnapmBusyError) as cm:
            self.manager.revert_snapshot_sets(selection)

        # Unmount the set, deactivate/reactivate and re-mount
        self.stop_start_storage()

    def test_rename_reverting_snapshot_set_raises(self):
        testset = "testset0"
        self.manager.create_snapshot_set(testset, self.mount_points())

        selection = snapm.Selection(name=testset)
        self.manager.revert_snapshot_sets(selection)

        with self.assertRaises(snapm.SnapmStateError) as cm:
            self.manager.rename_snapshot_set(testset, "testset1")

        # Unmount the set, deactivate/reactivate and re-mount
        self.stop_start_storage()

    def test_delete_reverting_snapshot_set_raises(self):
        testset = "testset0"
        self.manager.create_snapshot_set(testset, self.mount_points())

        selection = snapm.Selection(name=testset)
        self.manager.revert_snapshot_sets(selection)

        with self.assertRaises(snapm.SnapmBusyError) as cm:
            self.manager.delete_snapshot_sets(selection)

        # Unmount the set, deactivate/reactivate and re-mount
        self.stop_start_storage()

    def test_revert_snapshot_set_bad_name_raises(self):
        with self.assertRaises(snapm.SnapmNotFoundError) as cm:
            self.manager.revert_snapshot_set(name="nosuchset")

    def test_revert_snapshot_set_bad_uuid_raises(self):
        with self.assertRaises(snapm.SnapmNotFoundError) as cm:
            self.manager.revert_snapshot_set(uuid=UUID("00000000-0000-0000-0000-000000000000"))

    def test_revert_snapshot_set_name_uuid_conflict_raises(self):
        sset1 = self.manager.create_snapshot_set("testset0", self.mount_points())
        sset2 = self.manager.create_snapshot_set("testset1", self.mount_points())

        with self.assertRaises(snapm.SnapmInvalidIdentifierError) as cm:
            self.manager.revert_snapshot_set(name=sset1.name, uuid=sset2.uuid)

        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset1"))

    def test_resize_snapshot_set_bad_name_raises(self):
        with self.assertRaises(snapm.SnapmNotFoundError) as cm:
            self.manager.resize_snapshot_set([], name="nosuchset")

    def test_resize_snapshot_set_bad_uuid_raises(self):
        with self.assertRaises(snapm.SnapmNotFoundError) as cm:
            self.manager.resize_snapshot_set([], uuid=UUID("00000000-0000-0000-0000-000000000000"))

    def test_resize_snapshot_set_name_uuid_conflict_raises(self):
        sset1 = self.manager.create_snapshot_set("testset0", self.mount_points())
        sset2 = self.manager.create_snapshot_set("testset1", self.mount_points())

        with self.assertRaises(snapm.SnapmInvalidIdentifierError) as cm:
            self.manager.resize_snapshot_set([], name=sset1.name, uuid=sset2.uuid)

        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))
        self.manager.delete_snapshot_sets(snapm.Selection(name="testset1"))

    def test_resize_snapshot_set_non_member_raises(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())

        with self.assertRaises(snapm.SnapmNotFoundError) as cm:
            self.manager.resize_snapshot_set(["/home:1G"], name="testset0")

        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_resize_snapshot_set_no_space_raises(self):
        self.manager.create_snapshot_set("testset0", self.mount_points())

        with self.assertRaises(snapm.SnapmNoSpaceError) as cm:
            self.manager.resize_snapshot_set([], name="testset0", default_size_policy="20G")

        self.manager.delete_snapshot_sets(snapm.Selection(name="testset0"))

    def test_resize_snapshot_set_mount_specs(self):
        testset = "testset0"
        mount_specs = [f"{mp}:512MiB" for mp in self.mount_points()]
        self.manager.create_snapshot_set(testset, mount_specs)

        snapset = self.manager.find_snapshot_sets(snapm.Selection(name=testset))[0]
        for snapshot in snapset.snapshots:
            if snapshot.provider.name == "lvm2cow":
                self.assertEqual(snapshot.size, 512 * 1024 ** 2)

        resize_specs = [f"{self._lvm.mount_root}/{name}:1GiB" for name in self._lvm.volumes]
        self.manager.resize_snapshot_set(resize_specs, name=testset)

        snapset = self.manager.find_snapshot_sets(snapm.Selection(name=testset))[0]
        for snapshot in snapset.snapshots:
            if snapshot.provider.name == "lvm2cow":
                self.assertEqual(snapshot.size, 1024 ** 3)

        self.manager.delete_snapshot_sets(snapm.Selection(name=testset))

    def test_resize_snapshot_set_default_size_policy(self):
        testset = "testset0"
        mount_specs = [f"{mp}:512MiB" for mp in self.mount_points()]
        self.manager.create_snapshot_set(testset, mount_specs)

        snapset = self.manager.find_snapshot_sets(snapm.Selection(name=testset))[0]
        for snapshot in snapset.snapshots:
            if snapshot.provider.name == "lvm2cow":
                self.assertEqual(snapshot.size, 512 * 1024 ** 2)

        self.manager.resize_snapshot_set(None, name=testset, default_size_policy="1G")

        snapset = self.manager.find_snapshot_sets(snapm.Selection(name=testset))[0]
        for snapshot in snapset.snapshots:
            if snapshot.provider.name == "lvm2cow":
                self.assertEqual(snapshot.size, 1024 ** 3)

    def test_split_snapshot_set_split(self):
        testset = "testset0"
        splitset = "testset1"
        self.manager.create_snapshot_set(testset, self.mount_points())

        split_source = self.mount_points()[0]
        nosplit_sources = self.mount_points()[1:]

        # Split split_source from snapshot set
        split = self.manager.split_snapshot_set(testset, splitset, [split_source])

        # Get updated snapshot set
        test = self.manager.find_snapshot_sets(selection=snapm.Selection(name=testset))[0]

        self.assertEqual(split.name, splitset)
        self.assertTrue(split_source in split.sources)
        self.assertTrue(split_source not in test.sources)
        for source in nosplit_sources:
            self.assertTrue(source in test.sources)

    def test_split_snapshot_set_split_bad_name(self):
        testset = "testset0"
        splitset = "testset1"
        badset = "badset0"
        self.manager.create_snapshot_set(testset, self.mount_points())

        split_source = self.mount_points()[0]

        with self.assertRaises(snapm.SnapmNotFoundError):
            # Attempt to split with bad snapset name
            split = self.manager.split_snapshot_set(badset, splitset, [split_source])

    def test_split_snapshot_set_split_empty_split(self):
        testset = "testset0"
        splitset = "testset1"
        self.manager.create_snapshot_set(testset, self.mount_points())

        split_sources = self.mount_points()

        with self.assertRaises(snapm.SnapmArgumentError):
            # Split split_source from snapshot set
            split = self.manager.split_snapshot_set(testset, splitset, split_sources)

    def test_split_snapshot_set_split_size_policy_raises(self):
        testset = "testset0"
        splitset = "testset1"
        self.manager.create_snapshot_set(testset, self.mount_points())

        split_sources = [f"{mp}:10%SIZE" for mp in self.mount_points()]

        with self.assertRaises(snapm.SnapmArgumentError):
            # Attempt to split with size policies in source_specs
            split = self.manager.split_snapshot_set(testset, splitset, split_sources)

    def test_split_snapshot_set_split_bad_source_raises(self):
        testset = "testset0"
        splitset = "testset1"
        self.manager.create_snapshot_set(testset, self.mount_points())

        split_source = "/quux"

        with self.assertRaises(snapm.SnapmNotFoundError):
            # Split non-existent source from snapshot set
            split = self.manager.split_snapshot_set(testset, splitset, [split_source])

    def test_split_snapshot_set_split_empty_sources(self):
        testset = "testset0"
        splitset = "testset1"
        self.manager.create_snapshot_set(testset, self.mount_points())

        with self.assertRaises(snapm.SnapmArgumentError):
            # Split [] from snapshot set
            split = self.manager.split_snapshot_set(testset, splitset, [])

    def test_split_snapshot_set_prune(self):
        testset = "testset0"
        self.manager.create_snapshot_set(testset, self.mount_points())

        prune_source = self.mount_points()[0]
        noprune_sources = self.mount_points()[1:]

        # Prune prune_source from snapshot set
        prune = self.manager.split_snapshot_set(testset, None, [prune_source])

        self.assertTrue(prune_source not in prune.sources)
        for source in noprune_sources:
            self.assertTrue(source in prune.sources)

    def test_split_snapshot_set_prune_bad_name(self):
        testset = "testset0"
        badset = "badset0"
        self.manager.create_snapshot_set(testset, self.mount_points())

        split_source = self.mount_points()[0]

        with self.assertRaises(snapm.SnapmNotFoundError):
            # Attempt to split with bad snapset name
            split = self.manager.split_snapshot_set(badset, None, [split_source])

    def test_split_snapshot_set_prune_empty_prune(self):
        testset = "testset0"
        self.manager.create_snapshot_set(testset, self.mount_points())

        split_sources = self.mount_points()

        with self.assertRaises(snapm.SnapmArgumentError):
            # Split split_source from snapshot set
            split = self.manager.split_snapshot_set(testset, None, split_sources)

    def test_split_snapshot_set_prune_empty_sources(self):
        testset = "testset0"
        self.manager.create_snapshot_set(testset, self.mount_points())

        with self.assertRaises(snapm.SnapmArgumentError):
            # Split [] from snapshot set
            split = self.manager.split_snapshot_set(testset, None, [])

    def test_split_snapshot_set_prune_size_policy_raises(self):
        testset = "testset0"
        self.manager.create_snapshot_set(testset, self.mount_points())

        split_sources = [f"{mp}:10%SIZE" for mp in self.mount_points()]

        with self.assertRaises(snapm.SnapmArgumentError):
            # Attempt to split with size policies in source_specs
            split = self.manager.split_snapshot_set(testset, None, split_sources)

    def test_split_snapshot_set_prune_bad_source_raises(self):
        testset = "testset0"
        self.manager.create_snapshot_set(testset, self.mount_points())

        split_source = "/quux"

        with self.assertRaises(snapm.SnapmNotFoundError):
            # Split non-existent source from snapshot set
            split = self.manager.split_snapshot_set(testset, None, [split_source])


@unittest.skipIf(not have_root(), "requires root privileges")
class ManagerTestsThin(unittest.TestCase):
    """
    Tests for snapm.manager.Manager that apply only to thin provisioned
    snapshot providers (lvm2thin and stratis).
    """
    volumes = [] # Do not use lvm2cow
    thin_volumes = ["root", "home", "var", "data_vol"]
    stratis_volumes = ["fs1", "fs2"]

    def setUp(self):
        log.debug("Preparing %s", self._testMethodName)
        def cleanup_lvm():
            log.debug("Cleaning up LVM (%s)", self._testMethodName)
            if hasattr(self, "_lvm"):
                self._lvm.destroy()

        def cleanup_stratis():
            log.debug("Cleaning up Stratis (%s)", self._testMethodName)
            if hasattr(self, "_stratis"):
                self._stratis.destroy()

        self.addCleanup(cleanup_lvm)
        self.addCleanup(cleanup_stratis)

        self._lvm = LvmLoopBacked(self.volumes, thin_volumes=self.thin_volumes)
        self._stratis = StratisLoopBacked(self.stratis_volumes)

        self.manager = snapm.manager.Manager()

    def mount_points(self):
        return self._lvm.mount_points() + self._stratis.mount_points()

    def test_create_snapshot_set_recursion_raises(self):
        sset = self.manager.create_snapshot_set("testset0", self.mount_points())
        self.manager.activate_snapshot_sets(snapm.Selection(name="testset0"))
        snap_devs = [snapshot.devpath for snapshot in sset.snapshots]
        with self.assertRaises(snapm.SnapmRecursionError) as cm:
            sset = self.manager.create_snapshot_set("testset1", snap_devs)



# Copyright Red Hat
#
# snapm/manager/plugins/__init__.py - Snapshot Manager plugins
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: Apache-2.0
"""
Snapshot manager plugin helpers.
"""
import os
from os.path import sep as path_sep, ismount

from snapm import SNAPM_VALID_NAME_CHARS, Snapshot

_MOUNT_SEPARATOR = "-"
_ESCAPED_MOUNT_SEPARATOR = "--"

_PROC_MOUNTS = "/proc/self/mounts"

DEV_PREFIX = "/dev"
DEV_MAPPER_PREFIX = "/dev/mapper/"

# dmsetup options
DMSETUP_CMD = "dmsetup"
DMSETUP_SPLITNAME = "splitname"
DMSETUP_INFO = "info"
DMSETUP_LVM_SUBSYS = "LVM"
DMSETUP_NO_HEADINGS = "--noheadings"
DMSETUP_COLUMNS = "--columns"
DMSETUP_FIELDS_VG_LV = "-ovg_name,lv_name"
DMSETUP_FIELDS_UUID = "-ouuid"
DMSETUP_REPORT_SEP = ":"

# Fields: origin_name-snapset_snapset-name_timestamp_mount-point
SNAPSHOT_NAME_FORMAT = "%s-snapset_%s_%d_%s"


class Plugin:
    """
    Abstract base class for snapshot manager plugins.
    """

    name = "plugin"
    version = "0.1.0"
    snapshot_class = Snapshot

    def __init__(self, logger):
        self.size_map = None
        self.logger = logger

    def _log_error(self, *args):
        """
        Log at error level.
        """
        self.logger.error(*args)

    def _log_warn(self, *args):
        """
        Log at warning level.
        """
        self.logger.warning(*args)

    def _log_info(self, *args):
        """
        Log at info level.
        """
        self.logger.info(*args)

    def _log_debug(self, *args):
        """
        Log at debug level.
        """
        self.logger.debug(*args)

    def info(self):
        """
        Return plugin name and version.
        """
        return {"name": self.name, "version": self.version}

    def start_transaction(self):
        """
        Begin a snapshot set creation transaction in this plugin.
        """
        self.size_map = {}

    def end_transaction(self):
        """
        End a snapshot set creation transaction in this plugin.
        """
        self.size_map = None

    def discover_snapshots(self):
        """
        Discover snapshots managed by this plugin class.

        Returns a list of objects that are a subclass of ``Snapshot``.
        :returns: A list of snapshots discovered by this plugin class.
        """
        raise NotImplementedError

    def can_snapshot(self, source):
        """
        Test whether this plugin can snapshot the specified mount point or
        block device.

        :param source: The block device or mount point path to test.
        :returns: ``True`` if this plugin can snapshot the file system mounted
                  at ``mount_point``, or ``False`` otherwise.
        """
        raise NotImplementedError

    def check_create_snapshot(
        self, origin, snapset_name, timestamp, mount_point, size_policy
    ):
        """
        Perform pre-creation checks before creating a snapshot.

        :param origin: The origin volume for the snapshot.
        :param snapset_name: The name of the snapshot set to be created.
        :param timestamp: The snapshot set timestamp.
        :param mount_point: The mount point path for this snapshot.
        :raises: ``SnapmNoSpaceError`` if there is insufficient free space to
                 create the snapshot.
        """
        raise NotImplementedError

    def create_snapshot(
        self, origin, snapset_name, timestamp, mount_point, size_policy
    ):
        """
        Create a snapshot of ``origin`` in the snapset named ``snapset_name``.

        :param origin: The origin volume for the snapshot.
        :param snapset_name: The name of the snapshot set to be created.
        :param timestamp: The snapshot set timestamp.
        :param mount_point: The mount point path for this snapshot.
        :raises: ``SnapmNoSpaceError`` if there is insufficient free space to
                 create the snapshot.
        """
        raise NotImplementedError

    def rename_snapshot(self, old_name, origin, snapset_name, timestamp, mount_point):
        """
        Rename the snapshot named ``old_name`` according to the provided
        snapshot field values.

        :param old_name: The original name of the snapshot to be renamed.
        :param origin: The origin volume for the snapshot.
        :param snapset_name: The new name of the snapshot set.
        :param timestamp: The snapshot set timestamp.
        :param mount_point: The mount point of the snapshot.
        """
        raise NotImplementedError

    def check_resize_snapshot(self, name, origin, mount_point, size_policy):
        """
        Chcek whether this snapshot can be resized to the requested
        ``size_policy``. This method returns if the resize can be satisfied and
        raises an exception if not.

        :returns: None
        :raises: ``SnapmNoSpaceError`` if insufficient space is available to
        satisfy the requested size policy or ``SnapmPluginError`` if another
        reason prevents the snapshot from being resized.
        """
        raise NotImplementedError

    def resize_snapshot(self, name, origin, mount_point, size_policy):
        """
        Attempt to resize the snapshot ``name`` to the requested
        ``size_policy``. This method returns if the resize can be satisfied and
        raises an exception if not.

        :returns: None
        :raises: ``SnapmNoSpaceError`` if insufficient space is available to
        satisfy the requested size policy or ``SnapmPluginError`` if another
        reason prevents the snapshot from being resized.
        """
        raise NotImplementedError

    def check_revert_snapshot(self, name, origin):
        """
        Check whether this snapshot can be reverted or not. This method returns
        if the current snapshot can be reverted and raises an exception if not.

        :returns: None
        :raises: ``NotImplementedError`` if this plugin does not support the
        revert operation, ``SnapmBusyError`` if the snapshot is already in the
        process of being reverted to another snapshot state or
        ``SnapmPluginError`` if another reason prevents the snapshot from being
        merged.
        """
        raise NotImplementedError

    def revert_snapshot(self, name):
        """
        Request to revert a snapshot and revert the content of the origin
        volume to its state at the time of the snapshot.

        This may be deferred until the next device activation or mount
        operation for the respective volume.

        :param name: The name of the snapshot to revert.
        """
        raise NotImplementedError

    def delete_snapshot(self, name):
        """
        Delete the snapshot named ``name``

        :param name: The name of the snapshot to be removed.
        """
        raise NotImplementedError

    def activate_snapshot(self, name):
        """
        Activate the snapshot named ``name``

        :param name: The name of the snapshot to be activated.
        """
        raise NotImplementedError

    def deactivate_snapshot(self, name):
        """
        Deactivate the snapshot named ``name``

        :param name: The name of the snapshot to be deactivated.
        """
        raise NotImplementedError

    def set_autoactivate(self, name, auto=False):
        """
        Set the autoactivation state of the snapshot named ``name``.

        :param name: The name of the snapshot to be modified.
        :param auto: ``True`` to enable autoactivation or ``False`` otherwise.
        """
        raise NotImplementedError

    def origin_from_mount_point(self, mount_point):
        """
        Return a string representing the origin from a given mount point path.

        :param mount_point: The mount point path.
        """
        raise NotImplementedError


def _escape_bad_chars(path):
    """
    Encode illegal characters in mount point path.

    :param path: The path to escape.
    :returns: The escaped path.
    """
    escaped = ""
    for char in path:
        if char != path_sep and char not in SNAPM_VALID_NAME_CHARS:
            escaped = escaped + "." + char.encode("utf8").hex()
        else:
            if char == ".":
                escaped = escaped + ".."
            else:
                escaped = escaped + char
    return escaped


def _unescape_bad_chars(path):
    """
    Decode illegal characters in mount point path.

    :param path: The path to unescape.
    :returns: The unescaped path.
    """

    def decode_byte(byte):
        return bytearray.fromhex(byte).decode("utf8")

    unescaped = ""
    i = 0
    while i < len(path):
        if path[i] == ".":
            if path[i + 1] == ".":
                unescaped = unescaped + "."
                i += 1
            else:
                unescaped = unescaped + decode_byte(path[i + 1 : i + 3])
                i += 2
        else:
            unescaped = unescaped + path[i]
        i += 1
    return unescaped


def encode_mount_point(mount_point):
    """
    Encode mount point paths.

    Encode a mount point for use in a snapshot name by replacing the path
    separator with '-'.
    """
    if mount_point == "/":
        return _MOUNT_SEPARATOR
    mount_point = _escape_bad_chars(mount_point)
    parts = [
        part.replace(_MOUNT_SEPARATOR, _ESCAPED_MOUNT_SEPARATOR)
        for part in mount_point.split(path_sep)
    ]
    return _MOUNT_SEPARATOR.join(parts)


def _split_mount_separators(mount_str):
    """
    Split string by mount separator handling escaped separators.

    Split the mount string ``mount_str`` at ``_MOUNT_SEPARATOR`` boundaries,
    into path components, ignoring ``_ESCAPED_MOUNT_SEPARATOR``.
    """
    result = []
    current = ""
    i = 0
    while i < len(mount_str):
        if mount_str[i] == _MOUNT_SEPARATOR:
            if i < len(mount_str) - 1 and mount_str[i + 1] == _MOUNT_SEPARATOR:
                current += _ESCAPED_MOUNT_SEPARATOR
                i += 1
            else:
                result.append(current)
                current = ""
        else:
            current += mount_str[i]
        i += 1

    result.append(current)  # Append the remaining segment
    return result


def decode_mount_point(mount_str):
    """
    Parse the mount point component of a snapshot name.
    """
    mount_str = _unescape_bad_chars(mount_str)
    mount_parts = _split_mount_separators(mount_str)
    unescaped_parts = [
        part.replace(_ESCAPED_MOUNT_SEPARATOR, _MOUNT_SEPARATOR) for part in mount_parts
    ]
    return path_sep.join(unescaped_parts)


def parse_snapshot_name(full_name, origin):
    """
    Attempt to parse a snapshot set snapshot name.

    Returns a tuple of (snapset_name, timestamp, mount_point) if ``full_name``
    is a valid snapset snapshot name, or ``None`` otherwise.
    """
    if not full_name.startswith(origin):
        return None
    base = full_name.removeprefix(origin + "-")
    if "_" not in base:
        return None
    fields = base.split("_", maxsplit=3)
    if len(fields) != 4:
        return None
    if fields[0] != "snapset":
        return None
    snapset_name = fields[1]
    timestamp = int(fields[2])
    mount_point = decode_mount_point(fields[3])
    # (snapset_name, timestamp, mount)
    return (snapset_name, timestamp, mount_point)


def _parse_proc_mounts_line(mount_line):
    """
    Parse ``/proc/mounts`` line.

    Parse a line of ``/proc/mounts`` data and return a tuple
    containing the device and mount point.
    """
    (device, mount_point, _) = mount_line.split(maxsplit=2)
    return (device, mount_point)


def device_from_mount_point(mount_point):
    """
    Convert a mount point path to a corresponding device path.

    Return the device corresponding to the file system mount point
    ``mount_point`` according to /proc/self/mounts.
    """
    # Ignore any trailing '/' in mount point path
    if mount_point != path_sep:
        mount_point = mount_point.rstrip(path_sep)
    with open(_PROC_MOUNTS, "r", encoding="utf8") as mounts:
        for line in mounts.readlines():
            (device, mount_path) = _parse_proc_mounts_line(line)
            if mount_path == mount_point:
                return device
    raise KeyError(f"Mount point {mount_point} not found")


def mount_point_space_used(mount_point):
    """
    Retrieve space usage for a mount point path.

    Return the space used on the file system mounted at ``mount_point``
    as a value in bytes.
    """
    if not mount_point or not ismount(mount_point):
        return -1
    stat = os.statvfs(mount_point)
    return (stat.f_blocks - stat.f_bfree) * stat.f_frsize


def format_snapshot_name(origin, snapset_name, timestamp, mount_point):
    """
    Format structured snapshot name.

    Format a snapshot name according to the snapshot manager naming format.
    """
    return SNAPSHOT_NAME_FORMAT % (origin, snapset_name, timestamp, mount_point)


__all__ = [
    "DEV_PREFIX",
    "DEV_MAPPER_PREFIX",
    "DMSETUP_CMD",
    "DMSETUP_SPLITNAME",
    "DMSETUP_INFO",
    "DMSETUP_LVM_SUBSYS",
    "DMSETUP_NO_HEADINGS",
    "DMSETUP_COLUMNS",
    "DMSETUP_FIELDS_VG_LV",
    "DMSETUP_FIELDS_UUID",
    "DMSETUP_REPORT_SEP",
    "Plugin",
    "encode_mount_point",
    "decode_mount_point",
    "parse_snapshot_name",
    "device_from_mount_point",
    "mount_point_space_used",
    "format_snapshot_name",
]



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
XML interface specifications.
"""
# isort: STDLIB
import os
import xml.etree.ElementTree as ET  # nosec B405

# isort: FIRSTPARTY
from dbus_client_gen import (
    DbusClientGenerationError,
    managed_object_class,
    mo_query_builder,
)
from dbus_python_client_gen import DPClientGenerationError, make_class

from ._errors import StratisCliGenerationError
from ._constants import (
    BLOCKDEV_INTERFACE,
    FILESYSTEM_INTERFACE,
    MANAGER_INTERFACE,
    POOL_INTERFACE,
    REPORT_INTERFACE,
)
from ._environment import get_timeout
from ._introspect import SPECS

DBUS_TIMEOUT_SECONDS = 120


try:
    # pylint: disable=invalid-name

    timeout = get_timeout(
        os.environ.get("STRATIS_DBUS_TIMEOUT", DBUS_TIMEOUT_SECONDS * 1000)
    )

    report_spec = ET.fromstring(SPECS[REPORT_INTERFACE])  # nosec B314
    Report = make_class("Report", report_spec, timeout)

    filesystem_spec = ET.fromstring(SPECS[FILESYSTEM_INTERFACE])  # nosec B314
    Filesystem = make_class("Filesystem", filesystem_spec, timeout)
    MOFilesystem = managed_object_class("MOFilesystem", filesystem_spec)
    filesystems = mo_query_builder(filesystem_spec)

    pool_spec = ET.fromstring(SPECS[POOL_INTERFACE])  # nosec B314
    Pool = make_class("Pool", pool_spec, timeout)
    MOPool = managed_object_class("MOPool", pool_spec)
    pools = mo_query_builder(pool_spec)

    blockdev_spec = ET.fromstring(SPECS[BLOCKDEV_INTERFACE])  # nosec B314
    MODev = managed_object_class("MODev", blockdev_spec)
    devs = mo_query_builder(blockdev_spec)

    Manager = make_class(
        "Manager", ET.fromstring(SPECS[MANAGER_INTERFACE]), timeout  # nosec B314
    )

    ObjectManager = make_class(
        "ObjectManager",
        ET.fromstring(SPECS["org.freedesktop.DBus.ObjectManager"]),  # nosec B314
        timeout,
    )

    # Specification for the lowest manager interface supported by the major
    # version of stratisd on which this version of the CLI depends.
    # This minimal specification includes only the specification for the
    # Version property.
    manager_spec = """
    <interface name="org.storage.stratis3.Manager.r0">
        <property access="read" name="Version" type="s">
            <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="const" />
        </property>
    </interface>
    """
    Manager0 = make_class(
        "Manager0", ET.fromstring(manager_spec), timeout  # nosec B314
    )

# Do not expect to get coverage on Generation errors.
# These can only occurs if the XML data in _SPECS is ill-formed; we have
# complete control over that data and can expect it to be valid.
except DPClientGenerationError as err:  # pragma: no cover
    raise StratisCliGenerationError(
        "Failed to generate some class needed for invoking dbus-python methods"
    ) from err
except DbusClientGenerationError as err:  # pragma: no cover
    raise StratisCliGenerationError(
        "Failed to generate some class needed for examining D-Bus data"
    ) from err


def _add_abs_path_assertion(klass, method_name, key):
    """
    Set method_name of method_klass to a new method which checks that the
    device paths values at key are absolute paths.

    :param klass: the klass to which this metthod belongs
    :param str method_name: the name of the method
    :param str key: the key at which the paths can be found in the arguments
    """
    method_class = getattr(klass, "Methods")
    orig_method = getattr(method_class, method_name)

    def new_method(proxy, args):
        """
        New CreatePool method
        """
        rel_paths = [path for path in args[key] if not os.path.isabs(path)]
        assert (
            rel_paths == []
        ), f"Precondition violated: paths {', '.join(rel_paths)} should be absolute"
        return orig_method(proxy, args)

    setattr(method_class, method_name, new_method)


try:
    _add_abs_path_assertion(Manager, "CreatePool", "devices")
    _add_abs_path_assertion(Pool, "InitCache", "devices")
    _add_abs_path_assertion(Pool, "AddCacheDevs", "devices")
    _add_abs_path_assertion(Pool, "AddDataDevs", "devices")

except AttributeError as err:  # pragma: no cover
    # This can only happen if the expected method is missing from the XML spec
    # or code generation has a bug, we will never test for these conditions.
    raise StratisCliGenerationError(
        "Malformed class definition; could not access a class or method in "
        "the generated class definition"
    ) from err
